import numpy as np
import pandas as pd
from libpysal import weights
from sklearn.base import BaseEstimator

from esda.crand import _prepare_univariate
from esda.crand import crand as _crand_plus
from esda.crand import njit as _njit

PERMUTATIONS = 999


class Join_Counts_Local(BaseEstimator):  # noqa: N801
    """Univariate Local Join Count Statistic"""

    def __init__(
        self,
        connectivity=None,
        permutations=PERMUTATIONS,
        n_jobs=1,
        keep_simulations=True,
        seed=None,
        island_weight=0,
        drop_islands=True,
    ):
        """
        Initialize a Local_Join_Count estimator

        Parameters
        ----------
        connectivity : W | Graph
            spatial weights instance as W or Graph aligned with y
        permutations : int
            number of random permutations for calculation of pseudo
            p_values
        n_jobs : int
            Number of cores to be used in the conditional randomisation. If -1,
            all available cores are used.
        keep_simulations : bool (default True)
           If True, the entire matrix of replications under the null
           is stored in memory and accessible; otherwise, replications
           are not saved
        seed : None/int
           Seed to ensure reproducibility of conditional randomizations.
           Must be set here, and not outside of the function, since numba
           does not correctly interpret external seeds
           nor numpy.random.RandomState instances.
        island_weight:
            value to use as a weight for the "fake" neighbor for every island.
            If numpy.nan, will propagate to the final local statistic depending
            on the `stat_func`. If 0, then the lag is always zero for islands.
        drop_islands : bool (default True)
            Whether or not to preserve islands as entries in the adjacency
            list. By default, observations with no neighbors do not appear
            in the adjacency list. If islands are kept, they are coded as
            self-neighbors with zero weight. See ``libpysal.weights.to_adjlist()``.

        Attributes
        ----------
        LJC             : numpy array
                          array containing the univariate
                          Local Join Count (LJC).
        p_sim           : numpy array
                          array containing the simulated
                          p-values for each unit.

        """

        self.connectivity = connectivity
        self.permutations = permutations
        self.n_jobs = n_jobs
        self.keep_simulations = keep_simulations
        self.seed = seed
        self.island_weight = island_weight
        self.drop_islands = drop_islands

    def fit(self, y, n_jobs=1, permutations=999):
        """
        Parameters
        ----------
        y               : numpy.ndarray
                          array containing binary (0/1) data
        Returns
        -------
        the fitted estimator.

        Notes
        -----
        Technical details and derivations found in :cite:`AnselinLi2019`.

        Examples
        --------
        >>> import libpysal
        >>> w = libpysal.weights.lat2W(4, 4)
        >>> y = np.ones(16)
        >>> y[0:8] = 0
        >>> LJC_uni = Local_Join_Count(connectivity=w).fit(y)
        >>> LJC_uni.LJC
        >>> LJC_uni.p_sim

        Guerry data replicating GeoDa tutorial
        >>> import libpysal
        >>> import geopandas as gpd
        >>> guerry = libpysal.examples.load_example('Guerry')
        >>> guerry_ds = gpd.read_file(guerry.get_path('Guerry.shp'))
        >>> guerry_ds['SELECTED'] = 0
        >>> guerry_ds.loc[(guerry_ds['Donatns'] > 10997), 'SELECTED'] = 1
        >>> w = libpysal.weights.Queen.from_dataframe(guerry_ds)
        >>> LJC_uni = Local_Join_Count(connectivity=w).fit(guerry_ds['SELECTED'])
        >>> LJC_uni.LJC
        >>> LJC_uni.p_sim
        """
        # Need to ensure that the np.array() are of
        # dtype='float' for numba
        y = np.array(y, dtype="float")

        w = self.connectivity
        # Fill the diagonal with 0s
        if isinstance(w, weights.W):
            w = weights.util.fill_diagonal(w, val=0)
            w.transform = "b"
        else:
            w = w.assign_self_weight(0).eliminate_zeros().transform("b")

        keep_simulations = self.keep_simulations
        n_jobs = self.n_jobs
        # seed = self.seed

        self.y = y
        self.n = len(y)
        self.w = w

        self.LJC = self._statistic(y, w, self.drop_islands)

        if permutations:
            self.p_sim, self.rjoins = _crand_plus(
                z=self.y,
                w=self.w,
                observed=self.LJC,
                permutations=permutations,
                keep=keep_simulations,
                n_jobs=n_jobs,
                stat_func=_ljc_uni,
                island_weight=self.island_weight,
            )
            # Set p-values for those with LJC of 0 to NaN
            self.p_sim[self.LJC == 0] = "NaN"

        return self

    @staticmethod
    def _statistic(y, w, drop_islands):
        # Create adjacency list. Note that remove_symmetric=False - this is
        # different from the esda.Join_Counts() function.
        if isinstance(w, weights.W):
            adj_list = w.to_adjlist(remove_symmetric=False, drop_islands=drop_islands)
            zseries = pd.Series(y, index=w.id_order)
        else:
            if drop_islands:
                adj_list = w.adjacency.drop(w.isolates).reset_index()
            else:
                adj_list = w.adjacency.reset_index()
            zseries = pd.Series(y, index=w.unique_ids)
        focal = zseries.loc[adj_list.focal].values
        neighbor = zseries.loc[adj_list.neighbor].values
        LJC = (focal == 1) & (neighbor == 1)
        adj_list_LJC = pd.DataFrame(
            adj_list.focal.values, LJC.astype("uint8")
        ).reset_index()
        adj_list_LJC.columns = ["LJC", "ID"]
        adj_list_LJC = adj_list_LJC.groupby(by="ID").sum()
        LJC = np.array(adj_list_LJC.LJC.values, dtype="float")
        return LJC


# --------------------------------------------------------------
# Conditional Randomization Function Implementations
# --------------------------------------------------------------

# Note: scaling not used


@_njit(fastmath=True)
def _ljc_uni(i, z, permuted_ids, weights_i, scaling):  # noqa: ARG001
    # self_weight = weights_i[0]
    other_weights = weights_i[1:]
    zi, zrand = _prepare_univariate(i, z, permuted_ids, other_weights)
    return zi * (zrand @ other_weights)



# based off: https://github.com/pysal/esda/blob/master/tests/test_join_counts.py
import numpy as np
import pytest
from libpysal import graph
from libpysal.weights.util import lat2W

from esda.join_counts_local_bv import Join_Counts_Local_BV

parametrize_w = pytest.mark.parametrize(
    "w",
    [
        lat2W(4, 4),
        graph.Graph.from_W(lat2W(4, 4)),
    ],
    ids=["W", "Graph"],
)


class TestLocalJoinCountsBV:
    """Unit test for Local Join Counts (bivariate)"""

    def setup_method(self):
        self.x = np.ones(16)
        self.x[0:8] = 0
        self.z = [0, 1, 0, 1, 1, 1, 1, 1, 0, 0, 1, 1, 0, 0, 1, 1]

    @parametrize_w
    def test_bjc(self, w):
        np.random.seed(12345)
        ljc_bv_case1 = Join_Counts_Local_BV(connectivity=w).fit(
            self.x, self.z, case="BJC"
        )
        assert np.array_equal(
            ljc_bv_case1.LJC, [0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 0, 0, 0, 0, 0, 0]
        )

    @parametrize_w
    def test_clc(self, w):
        np.random.seed(12345)
        ljc_bv_case2 = Join_Counts_Local_BV(connectivity=w).fit(
            self.x, self.z, case="CLC"
        )
        assert np.array_equal(
            ljc_bv_case2.LJC, [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 2, 0, 0, 2, 2]
        )



import libpysal
import numpy as np
import pytest

from ..gamma import Gamma

parametrize_lat = pytest.mark.parametrize(
    "w",
    [
        libpysal.weights.util.lat2W(4, 4),
        libpysal.graph.Graph.from_W(libpysal.weights.util.lat2W(4, 4)),
    ],
    ids=["W", "Graph"],
)


class TestGamma:
    """Unit test for Gamma Index"""

    def setup_method(self):
        self.y = np.ones(16)
        self.y[0:8] = 0
        np.random.seed(12345)

    @parametrize_lat
    def test_default(self, w):
        g = Gamma(self.y, w)

        np.testing.assert_allclose(g.g, 20.0)
        np.testing.assert_allclose(g.g_z, 3.1879280354548638)
        np.testing.assert_allclose(g.p_sim_g, 0.0030000000000000001)
        np.testing.assert_allclose(g.min_g, 0.0)
        np.testing.assert_allclose(g.max_g, 20.0)
        np.testing.assert_allclose(g.mean_g, 11.093093093093094)

    @parametrize_lat
    def test_s(self, w):
        np.random.seed(12345)
        g1 = Gamma(self.y, w, operation="s")
        np.testing.assert_allclose(g1.g, 8.0)
        np.testing.assert_allclose(g1.g_z, -3.7057554345954791)
        np.testing.assert_allclose(g1.p_sim_g, 0.001)
        np.testing.assert_allclose(g1.min_g, 14.0)
        np.testing.assert_allclose(g1.max_g, 48.0)
        np.testing.assert_allclose(g1.mean_g, 25.623623623623622)

    @parametrize_lat
    def test_a(self, w):
        np.random.seed(12345)
        g2 = Gamma(self.y, w, operation="a")
        np.testing.assert_allclose(g2.g, 8.0)
        np.testing.assert_allclose(g2.g_z, -3.7057554345954791)
        np.testing.assert_allclose(g2.p_sim_g, 0.001)
        np.testing.assert_allclose(g2.min_g, 14.0)
        np.testing.assert_allclose(g2.max_g, 48.0)
        np.testing.assert_allclose(g2.mean_g, 25.623623623623622)

    @parametrize_lat
    def test_standardize(self, w):
        np.random.seed(12345)
        g3 = Gamma(self.y, w, standardize=True)
        np.testing.assert_allclose(g3.g, 32.0)
        np.testing.assert_allclose(g3.g_z, 3.7057554345954791)
        np.testing.assert_allclose(g3.p_sim_g, 0.001)
        np.testing.assert_allclose(g3.min_g, -48.0)
        np.testing.assert_allclose(g3.max_g, 20.0)
        np.testing.assert_allclose(g3.mean_g, -3.2472472472472473)

    @parametrize_lat
    def test_op(self, w):
        np.random.seed(12345)
        if isinstance(w, libpysal.graph.Graph):
            pytest.skip("Calleble not supported with Graph")

        def func(z, i, j):
            q = z[i] * z[j]
            return q

        g4 = Gamma(self.y, w, operation=func)
        np.testing.assert_allclose(g4.g, 20.0)
        np.testing.assert_allclose(g4.g_z, 3.1879280354548638)
        np.testing.assert_allclose(g4.p_sim_g, 0.0030000000000000001)

    @parametrize_lat
    def test_by_col(self, w):
        import pandas as pd

        g = Gamma(self.y, w)

        df = pd.DataFrame(self.y, columns=["y"])
        r1 = Gamma.by_col(df, ["y"], w=w)
        assert "y_gamma" in r1.columns
        assert "y_p_sim" in r1.columns
        this_gamma = np.unique(r1.y_gamma.values)
        this_pval = np.unique(r1.y_p_sim.values)

        np.testing.assert_allclose(this_gamma, g.g)
        np.testing.assert_allclose(this_pval, g.p_sim)
        Gamma.by_col(df, ["y"], inplace=True, operation="s", w=w)
        this_gamma = np.unique(df.y_gamma.values)
        this_pval = np.unique(df.y_p_sim.values)
        np.testing.assert_allclose(this_gamma, 8.0)
        np.testing.assert_allclose(this_pval, 0.001)



import geopandas
import numpy
import pytest
import shapely
from numpy import array, testing

import esda

pytest.importorskip("numba")


shape = array(
    [
        shapely.geometry.Polygon(
            [
                (0, 0),
                (0.25, 0.25),
                (0, 0.5),
                (0.25, 0.75),
                (0, 1),
                (1.25, 1),
                (0.75, 0.5),
                (1.25, 0),
            ]
        )
    ]
)

ATOL = 0.001

## for a hole/multipart testbench:

test_geom_translated = shapely.from_wkt(
    [
        "POLYGON (("
        "-3.1823503126754247 0.085191513232644, "
        "-3.2545854200972997 0.271135116748269, "
        "-3.2472001661910497 0.296769882373269, "
        "-3.2779008497847997 0.3333146821779565, "
        "-3.286461030448862 0.4668824922365502, "
        "-3.312919770683237 0.4887788545412377, "
        "-3.308738862480112 0.5528352510256127, "
        "-3.271751557792612 0.6005037080568627, "
        "-3.2749711622847997 0.6791780244631127, "
        "-3.301475678886362 0.7266938936037377, "
        "-3.3279496779097997 0.7266252290529565, "
        "-3.3439103712691747 0.8217256318849877, "
        "-3.385307465995737 0.9057634126467065, "
        "-3.385490571464487 1.0868776094240502, "
        "-3.4014665236129247 1.1563432466310815, "
        "-3.3219682325972997 1.1584108125490502, "
        "-3.3168107618941747 1.328538681201394, "
        "-3.2181779493941747 1.3360688936037377, "
        "-3.2150346388472997 1.4000871431154565, "
        "-3.1089555372847997 1.4057023774904565, "
        "-3.105888520683237 1.9232576143068627, "
        "-2.702140962089487 1.9063203584474877, "
        "-2.7046891798629247 0.8271501313967065, "
        "-2.749656831230112 0.7665956270021752, "
        "-2.671684419120737 0.5808656465334252, "
        "-2.8658219923629247 0.3313920747560815, "
        "-2.9108354200972997 0.1093156587404565, "
        "-3.1823503126754247 0.085191513232644"
        "))"
    ]
)[0]

test_simple = shapely.difference(
    shapely.box(-1, 0, -2, 1), shapely.box(-1, 0, -1.5, 0.5)
)

test_mp = shapely.union(shapely.box(-1, -1, -1.5, -2), shapely.box(0, -1, 1.25, -2))

test_hole = shapely.difference(
    shapely.box(0, 0, 1.81, 1.81), shapely.box(0.8, 0.8, 1.6, 1.6)
)

test_mp_hole = shapely.union(
    shapely.transform(
        test_hole, lambda x: numpy.column_stack((-x[:, 0] + 3, x[:, 1] * 0.5 + 3))
    ),
    shapely.transform(test_hole, lambda x: numpy.column_stack((x[:, 0] + 4, x[:, 1]))),
)

testbench = geopandas.GeoDataFrame(
    geometry=[test_geom_translated, test_simple, test_mp, test_hole, test_mp_hole]
).reset_index()
testbench["name"] = ["Hanock County", "Simple", "Multi", "Single Hole", "Multi Hole"]

# Create grid for testing of moment_of_inertia_regions
tmp_geoms = []
for i in range(8):
    for j in range(8):
        tmp_geoms.append(shapely.box(i, j, i+1, j+1))
tmp_geoseries = geopandas.GeoSeries(tmp_geoms)
regions = [1, 1, 1, 2, 2, 2, 3, 3,
           1, 1, 4, 4, 2, 2, 3, 3,
           1, 4, 4, 4, 4, 2, 3, 3,
           4, 4, 4, 4, 4, 4, 3, 3,
           4, 4, 4, 4, 4, 4, 3, 3,
           2, 4, 4, 4, 4, 2, 3, 3,
           2, 2, 4, 4, 2, 2, 3, 3,
           2, 2, 2, 2, 2, 2, 3, 3]
region_lookup = {
    1: "1 - Triangle region",
    2: "2 - Discontiguous region",
    3: "3 - Skinny region",
    4: "4 - Circle-ish region"
}
region_names = [region_lookup[i] for i in regions]
masses = [10, 20, 30, 40, 50] * 13
masses = masses[:-1]
test_region_8x8 = geopandas.GeoDataFrame(
    {
        "region_name": region_names,
        "mass": masses,
        "geometry": tmp_geoms
    }

)

def test_boundary_amplitude():
    observed = esda.shape.boundary_amplitude(shape)
    testing.assert_allclose(observed, 0.844527, atol=ATOL)


def test_convex_hull_ratio():
    observed = esda.shape.convex_hull_ratio(shape)
    testing.assert_allclose(observed, 0.7, atol=ATOL)


def test_length_width_diff():
    observed = esda.shape.length_width_diff(shape)
    testing.assert_allclose(observed, -0.25, atol=ATOL)


def test_radii_ratio():
    observed = esda.shape.radii_ratio(shape)
    testing.assert_allclose(observed, 0.659366, atol=ATOL)


def test_diameter_ratio():
    observed = esda.shape.diameter_ratio(shape)
    testing.assert_allclose(observed, 0.8, atol=ATOL)


def test_iaq():
    observed = esda.shape.isoareal_quotient(shape)
    testing.assert_allclose(observed, 0.622314, atol=ATOL)


def test_ipq():
    observed = esda.shape.isoperimetric_quotient(shape)
    testing.assert_allclose(observed, 0.387275, atol=ATOL)


def test_moment_of_inertia():
    observed = esda.shape.moment_of_inertia(testbench.geometry)
    testing.assert_allclose(
        observed,
        [0.23480628, 0.11458333, 1.57459077, 1.58210246, 14.18946959],
        atol=ATOL,
    )

def test_moment_of_inertia_normalized():
    observed = esda.shape.moment_of_inertia(testbench.geometry, normalize=True)
    testing.assert_allclose(
        observed,
        [0.58666692, 0.78130608, 0.30954837, 0.69905169, 0.175372  ],
        atol=ATOL,
    )

def test_moment_of_inertia_about_origin():
    observed = esda.shape.moment_of_inertia(testbench.geometry, ref_pt=(0,0))
    testing.assert_allclose(
        observed,
        [ 9.53818256,  2.25,        5.52604167,  5.24375414, 87.06458556],
        atol=ATOL,
    )

def test_moment_of_inertia_global():
    observed = esda.shape.moment_of_inertia_global(testbench)
    testing.assert_allclose(observed, 84.31724299486667, atol=ATOL)

# mir refers to `moment_of_inertia_regions`
def test_mir_with_regions():
    observed = numpy.asarray(esda.shape.moment_of_inertia_regions(test_region_8x8, regions=regions))
    testing.assert_allclose(
        observed,
        [  7.66666667, 205.22222222,  90.66666667,  96.        ],
        atol=ATOL,
    )

def test_mir_with_region_col():
    observed = numpy.asarray(esda.shape.moment_of_inertia_regions(test_region_8x8, regions="region_name"))
    testing.assert_allclose(
        observed,
        [  7.66666667, 205.22222222,  90.66666667,  96.        ],
        atol=ATOL,
    )

def test_mir_with_regions_normalized():
    observed = numpy.asarray(esda.shape.moment_of_inertia_regions(test_region_8x8, normalize=True, regions=regions))
    testing.assert_allclose(
        observed,
        [0.74733625, 0.25127007, 0.44937866, 0.95492966],
        atol=ATOL,
    )

def test_mir_with_weights():
    observed = numpy.asarray(esda.shape.moment_of_inertia_regions(test_region_8x8, weights=masses))
    testing.assert_allclose(
        observed,
        [1.66666667, 3.33333333, 5.        , 6.66666667, 8.33333333,
         1.66666667, 3.33333333, 5.        , 6.66666667, 8.33333333,
         1.66666667, 3.33333333, 5.        , 6.66666667, 8.33333333,
         1.66666667, 3.33333333, 5.        , 6.66666667, 8.33333333,
         1.66666667, 3.33333333, 5.        , 6.66666667, 8.33333333,
         1.66666667, 3.33333333, 5.        , 6.66666667, 8.33333333,
         1.66666667, 3.33333333, 5.        , 6.66666667, 8.33333333,
         1.66666667, 3.33333333, 5.        , 6.66666667, 8.33333333,
         1.66666667, 3.33333333, 5.        , 6.66666667, 8.33333333,
         1.66666667, 3.33333333, 5.        , 6.66666667, 8.33333333,
         1.66666667, 3.33333333, 5.        , 6.66666667, 8.33333333,
         1.66666667, 3.33333333, 5.        , 6.66666667, 8.33333333,
         1.66666667, 3.33333333, 5.        , 6.66666667],
        atol=ATOL,
    )

def test_mir_with_weight_col():
    observed = numpy.asarray(esda.shape.moment_of_inertia_regions(test_region_8x8, weights="mass"))
    testing.assert_allclose(
        observed,
        [1.66666667, 3.33333333, 5.        , 6.66666667, 8.33333333,
         1.66666667, 3.33333333, 5.        , 6.66666667, 8.33333333,
         1.66666667, 3.33333333, 5.        , 6.66666667, 8.33333333,
         1.66666667, 3.33333333, 5.        , 6.66666667, 8.33333333,
         1.66666667, 3.33333333, 5.        , 6.66666667, 8.33333333,
         1.66666667, 3.33333333, 5.        , 6.66666667, 8.33333333,
         1.66666667, 3.33333333, 5.        , 6.66666667, 8.33333333,
         1.66666667, 3.33333333, 5.        , 6.66666667, 8.33333333,
         1.66666667, 3.33333333, 5.        , 6.66666667, 8.33333333,
         1.66666667, 3.33333333, 5.        , 6.66666667, 8.33333333,
         1.66666667, 3.33333333, 5.        , 6.66666667, 8.33333333,
         1.66666667, 3.33333333, 5.        , 6.66666667, 8.33333333,
         1.66666667, 3.33333333, 5.        , 6.66666667],
        atol=ATOL,
    )

def test_mir_with_weights_normalized():
    observed = numpy.asarray(esda.shape.moment_of_inertia_regions(test_region_8x8, normalize=True, weights=masses))
    testing.assert_allclose(
        observed,
        esda.shape.moment_of_inertia(test_region_8x8, normalize=True),
        atol=ATOL,
    )

def test_mir_with_regions_and_weights():
    observed = numpy.asarray(esda.shape.moment_of_inertia_regions(test_region_8x8, regions=regions, weights=masses))
    testing.assert_allclose(
        observed,
        [ 189.50980392, 6185.74074074, 2709.58333333, 2771.57276995],
        atol=ATOL,
    )
def test_mir_with_regions_and_ref_pts():
    ref_pt = {1: (2, 0.5), 2: (1, 5), 3: (7, 7), 4: (5, 2)}
    observed = numpy.asarray(esda.shape.moment_of_inertia_regions(test_region_8x8, ref_pt=ref_pt, regions=regions))
    testing.assert_allclose(
        observed,
        [ 14.5       , 520.        , 234.66666667, 144.        ],
        atol=ATOL,
    )

def test_second_areal_moment():
    observed = esda.shape.second_areal_moment(testbench.geometry)
    testing.assert_allclose(
        observed,
        [0.23480628, 0.11458333, 1.57459077, 1.58210246, 14.18946959],
        atol=ATOL,
    )


def test_moa():
    observed = esda.shape.moa_ratio(shape)
    testing.assert_allclose(observed, 5.35261, atol=ATOL)


def test_nmi():
    observed = esda.shape.nmi(shape)
    testing.assert_allclose(observed, 0.802796, atol=ATOL)


def test_mbc():
    observed = esda.shape.minimum_bounding_circle_ratio(shape)
    testing.assert_allclose(observed, 0.434765, atol=ATOL)


def test_reflexive_angle_ratio():
    observed = esda.shape.reflexive_angle_ratio(shape)
    testing.assert_allclose(observed, 3 / 8, atol=ATOL)


def test_fractal_dimension():
    r = [
        esda.shape.fractal_dimension(shape, support=support)[0]
        for support in ("hex", "square", "circle")
    ]
    testing.assert_allclose(r, [0.218144, -4.29504, 0.257882], atol=ATOL)


def test_squareness():
    observed = esda.shape.squareness(shape)
    testing.assert_allclose(observed, 0.493094, atol=ATOL)


def test_rectangularity():
    observed = esda.shape.rectangularity(shape)
    testing.assert_allclose(observed, 0.7, atol=ATOL)


def test_shape_index():
    observed = esda.shape.shape_index(shape)
    testing.assert_allclose(observed, 0.659366, atol=ATOL)


def test_equivalent_rectangular_index():
    observed = esda.shape.equivalent_rectangular_index(shape)
    testing.assert_allclose(observed, 0.706581, atol=ATOL)


def test_form_factor():
    observed = esda.shape.form_factor(shape, array([2]))
    testing.assert_allclose(observed, 0.602535, atol=ATOL)



import numpy as np
import warnings

try:
    from numba import njit
except (ImportError, ModuleNotFoundError):
    from libpysal.common import jit as njit


def calculate_significance(test_stat, reference_distribution, alternative="two-sided"):
    """
    Calculate a pseudo p-value from a reference distribution.

    Pseudo-p values are calculated using the formula (M + 1) / (R + 1). Where R is the number of simulations
    and M is the number of times that the simulated value was equal to, or more extreme than the observed test statistic.

    Parameters
    ----------
    test_stat: float or numpy.ndarray
        The observed test statistic, or a vector of observed test statistics
    reference_distribution: numpy.ndarray
        A numpy array containing simulated test statistics as a result of conditional permutation.
    alternative: string
        One of 'two-sided', 'lesser', 'greater', 'folded', or 'directed'. Indicates the alternative hypothesis.
        - 'two-sided': the observed test statistic is in either tail of the reference distribution. This is an un-directed alternative hypothesis.
        - 'folded': the observed test statistic is an extreme value of the reference distribution folded about its mean. This is an un-directed alternative hypothesis.
        - 'lesser': the observed test statistic is small relative to the reference distribution. This is a directed alternative hypothesis.
        - 'greater': the observed test statistic is large relative to the reference distribution. This is a directed alternative hypothesis. 
        - 'directed': the observed test statistic is in either tail of the reference distribution, but the tail is selected depending on the test statistic. This is a directed alternative hypothesis, but the direction is chosen dependent on the data. This is not advised, and included solely to reproduce past results.

    Notes
    -----

    the directed p-value is half of the two-sided p-value, and corresponds to running the
    lesser and greater tests, then picking the smaller significance value. This is not advised, 
    since the p-value will be uniformly too small. 
    """
    reference_distribution = np.atleast_2d(reference_distribution)
    n_samples, p_permutations = reference_distribution.shape
    test_stat = np.atleast_2d(test_stat).reshape(n_samples, -1)
    if alternative not in (
        'folded',
        'two-sided',
        'greater',
        'lesser',
        'directed'
    ):
        raise ValueError(
            f"alternative='{alternative}' provided, but is not"
            f" one of the supported options: 'two-sided', 'greater', 'lesser', 'directed', 'folded')"
            )
    result = _permutation_significance(
        test_stat, 
        reference_distribution,
        alternative=alternative
        )
    if test_stat.size == 1:
        return result.item()
    else:
        return result

@njit(parallel=False, fastmath=False)
def _permutation_significance(test_stat, reference_distribution, alternative='two-sided'):
    reference_distribution = np.atleast_2d(reference_distribution)
    n_samples, p_permutations = reference_distribution.shape
    if isinstance(test_stat, (int, float)):
        test_stat = np.ones((n_samples,))*test_stat
    if alternative == "directed":
        larger = (reference_distribution >= test_stat).sum(axis=1)
        low_extreme = (p_permutations - larger) < larger
        larger[low_extreme] = p_permutations - larger[low_extreme]
        p_value = (larger + 1.0) / (p_permutations + 1.0)
    elif alternative == "lesser":
        p_value = (np.sum(reference_distribution <= test_stat, axis=1) + 1) / (
            p_permutations + 1
        )
    elif alternative == "greater":
        p_value = (np.sum(reference_distribution >= test_stat, axis=1) + 1) / (
            p_permutations + 1
        )
    elif alternative == "two-sided":
        # find percentile p at which the test statistic sits
        # find "synthetic" test statistic at 1-p
        # count how many observations are outisde of (p, 1-p)
        # including the test statistic and its synthetic pair
        lows = np.empty(n_samples).astype(reference_distribution.dtype)
        highs = np.empty(n_samples).astype(reference_distribution.dtype)
        for i in range(n_samples):
            percentile_i = (reference_distribution[i] <= test_stat[i]).mean()*100
            p_low = np.minimum(percentile_i, 100-percentile_i)
            lows[i] = np.percentile(
                reference_distribution[i], 
                p_low
            )
            highs[i] = np.percentile(
                reference_distribution[i], 
                100 - p_low
            )
        n_outside = (reference_distribution <= lows[:,None]).sum(axis=1)
        n_outside += (reference_distribution >= highs[:,None]).sum(axis=1)
        p_value = (n_outside + 1) / (p_permutations + 1)
    elif alternative == "folded":
        means = np.empty((n_samples,1)).astype(reference_distribution.dtype)
        for i in range(n_samples):
            means[i] = reference_distribution[i].mean()
        folded_test_stat = np.abs(test_stat - means)
        folded_reference_distribution = np.abs(reference_distribution - means)
        p_value = ((folded_reference_distribution >= folded_test_stat).sum(axis=1) + 1) / (
            p_permutations + 1
    	) 
    else:
        p_value = np.ones((n_samples, ))*np.nan
    return p_value





import libpysal
import numpy
import pytest
from sklearn.metrics import pairwise

from .. import silhouettes

RTOL = libpysal.common.RTOL
ATOL = libpysal.common.ATOL

parametrize_w = pytest.mark.parametrize(
    "w",
    [
        libpysal.weights.util.lat2W(3, 3, rook=False),
        libpysal.graph.Graph.from_W(libpysal.weights.util.lat2W(3, 3, rook=False)),
    ],
    ids=["W", "Graph"],
)


class TestSilhouette:
    def setup_method(self):
        numpy.random.seed(12345)
        self.X = numpy.random.random((9, 3))
        self.groups = numpy.array([[0, 0, 0], [0, 1, 1], [0, 1, 1]]).flatten()
        self.precomputed = self.X @ self.X.T
        self.altmetric = pairwise.manhattan_distances

    @parametrize_w
    def test_boundary(self, w):
        known = numpy.array(
            [
                0.03042804,
                -0.26071434,
                0.26664155,
                -0.09395997,
                0.02754655,
                -0.30581229,
                0.2697089,
                -0.29993007,
                0.0,
            ]
        )
        test = silhouettes.boundary_silhouette(self.X, self.groups, w)
        numpy.testing.assert_allclose(known, test, rtol=RTOL, atol=ATOL)
        known = numpy.array(
            [
                0.04246261,
                -0.17059996,
                0.35656939,
                -0.10765454,
                0.00831288,
                -0.30924291,
                0.36502035,
                -0.25464692,
                0.0,
            ]
        )
        test = silhouettes.boundary_silhouette(
            self.X, self.groups, w, metric=self.altmetric
        )
        numpy.testing.assert_allclose(known, test, rtol=RTOL, atol=ATOL)
        known = numpy.array(
            [
                -0.43402484,
                -0.29488836,
                -0.3663925,
                -0.32831205,
                -0.21536904,
                -0.10134236,
                -0.37144783,
                -0.01618446,
                0.0,
            ]
        )
        test = silhouettes.boundary_silhouette(
            self.X, self.groups, w, metric=self.precomputed
        )
        numpy.testing.assert_allclose(known, test, rtol=RTOL, atol=ATOL)
        with pytest.raises(AssertionError):
            silhouettes.boundary_silhouette(
                self.X,
                self.groups,
                w,
                metric=self.precomputed - self.precomputed.mean(),
            )

    @parametrize_w
    def test_path(self, w):
        known = numpy.array(
            [
                0.15982274,
                -0.02136909,
                -0.3972349,
                0.24479121,
                0.02754655,
                0.28465546,
                -0.07572727,
                0.26903733,
                0.4165144,
            ]
        )
        test = silhouettes.path_silhouette(self.X, self.groups, w)
        numpy.testing.assert_allclose(known, test, rtol=RTOL, atol=ATOL)
        known = numpy.array(
            [
                0.1520476,
                0.0390323,
                -0.34269345,
                0.27239358,
                0.00831288,
                0.27934432,
                -0.03874118,
                0.28623703,
                0.40062121,
            ]
        )
        test = silhouettes.path_silhouette(
            self.X, self.groups, w, metric=self.altmetric
        )
        numpy.testing.assert_allclose(known, test, rtol=RTOL, atol=ATOL)
        with pytest.raises(TypeError):
            silhouettes.path_silhouette(self.X, self.groups, w, metric=self.precomputed)
        with pytest.raises(AssertionError):
            silhouettes.path_silhouette(
                self.X, self.groups, w, metric=lambda d: -self.altmetric(d)
            )

    def test_nearest_label(self):
        known = numpy.array([1, 1, 1, 1, 0, 0, 1, 0, 0])
        test, d = silhouettes.nearest_label(self.X, self.groups, return_distance=True)
        numpy.testing.assert_array_equal(known, test)
        known = numpy.array([0, 1, 0, 0, 1, 0, 0, 0, 1])
        test = silhouettes.nearest_label(self.X, self.groups, keep_self=True)
        numpy.testing.assert_array_equal(test, known)
        knownd = numpy.array(
            [
                1.05707684,
                0.74780721,
                0.88841079,
                0.71628677,
                1.25964181,
                0.5854757,
                0.89710073,
                0.64575898,
                0.73913526,
            ]
        )
        numpy.testing.assert_allclose(d, knownd, rtol=RTOL, atol=ATOL)

    @parametrize_w
    def test_silhouette_alist(self, w):
        known = numpy.array(
            [
                0.0,
                0.0,
                0.22434244,
                0.0,
                0.0,
                0.0,
                -0.07589293,
                -0.07589293,
                0.0,
                0.41331324,
                0.41331324,
                0.0,
                0.0,
                0.11703681,
                0.0,
                0.11703681,
                0.27065991,
                0.27065991,
                0.27065991,
                0.27065991,
                0.0,
                0.27065991,
                0.0,
                0.0,
                -0.07441639,
                -0.07441639,
                0.0,
                0.0,
                0.0,
                0.0,
                0.41576712,
                0.41576712,
                -0.06657343,
                0.0,
                0.0,
                -0.06657343,
                0.0,
                0.0,
                0.0,
                0.0,
            ]
        )
        if isinstance(w, libpysal.weights.W):
            alist = w.to_adjlist(drop_islands=True)
        else:
            alist = w.adjacency.drop(w.isolates).reset_index()
        test = silhouettes.silhouette_alist(self.X, self.groups, alist)
        numpy.testing.assert_allclose(known, test.silhouette, rtol=RTOL, atol=ATOL)



import libpysal
import numpy as np
import pytest
from libpysal.common import ATOL, RTOL, pandas

from ..moran import Moran
from ..smaup import Smaup

PANDAS_EXTINCT = pandas is None


class TestSmaup:
    def setup_method(self):
        self.w = libpysal.io.open(libpysal.examples.get_path("stl.gal")).read()
        f = libpysal.io.open(libpysal.examples.get_path("stl_hom.txt"))
        self.y = np.array(f.by_col["HR8893"])
        self.rho = Moran(self.y, self.w).I
        self.n = len(self.y)
        self.k = int(self.n / 2)

    def test_defaults(self):
        sm = Smaup(self.n, self.k, self.rho)
        assert sm.n == 78
        assert sm.k == 39
        assert pytest.approx(sm.rho) == 0.24365582621771695
        np.testing.assert_allclose(sm.smaup, 0.15221341690376405, rtol=RTOL, atol=ATOL)
        assert pytest.approx(sm.critical_01) == 0.38970613333333337
        assert pytest.approx(sm.critical_05) == 0.3557221333333333
        assert pytest.approx(sm.critical_1) == 0.3157950666666666
        assert sm.summary == "Pseudo p-value > 0.10 (H0 is not rejected)"

    def test_sids(self):
        w = libpysal.io.open(libpysal.examples.get_path("sids2.gal")).read()
        f = libpysal.io.open(libpysal.examples.get_path("sids2.dbf"))
        SIDR = np.array(f.by_col("SIDR74"))
        rho = Moran(SIDR, w, two_tailed=False).I
        n = len(SIDR)
        k = int(n / 2)
        sm = Smaup(n, k, rho)
        np.testing.assert_allclose(sm.smaup, 0.15176796553181948, rtol=RTOL, atol=ATOL)
        assert pytest.approx(sm.critical_01) == 0.23404000000000003
        assert pytest.approx(sm.critical_05) == 0.21088
        assert pytest.approx(sm.critical_1) == 0.18239
        assert sm.summary == "Pseudo p-value > 0.10 (H0 is not rejected)"

    @pytest.mark.skipif(PANDAS_EXTINCT, reason="missing pandas")
    def test_by_col(self):
        from libpysal.io import geotable as pdio

        np.random.seed(11213)
        df = pdio.read_files(libpysal.examples.get_path("sids2.dbf"))
        w = libpysal.io.open(libpysal.examples.get_path("sids2.gal")).read()
        k = int(w.n / 2)
        mi = Moran.by_col(df, ["SIDR74"], w=w, two_tailed=False)
        rho = np.unique(mi.SIDR74_moran.values).item()
        sm = Smaup(w.n, k, rho)
        np.testing.assert_allclose(sm.smaup, 0.15176796553181948, atol=ATOL, rtol=RTOL)
        assert pytest.approx(sm.critical_01) == 0.23404000000000003
        assert pytest.approx(sm.critical_05) == 0.21088
        assert pytest.approx(sm.critical_1) == 0.18239
        assert sm.summary == "Pseudo p-value > 0.10 (H0 is not rejected)"



import warnings

import numpy as np
from libpysal import weights
from scipy import sparse as sp
from scipy.sparse import csgraph as cg

try:
    import pandas as pd
    import sklearn.metrics as sk
    import sklearn.metrics.pairwise as skp
    from sklearn.preprocessing import LabelEncoder  # noqa: F401

    HAS_REQUIREMENTS = True
except ImportError:
    HAS_REQUIREMENTS = False


def _raise_initial_error():
    missing = []
    try:
        import sklearn  # noqa: F401
    except ImportError:
        missing.append("scikit-learn")
    try:
        import pandas  # noqa: F401
    except ImportError:
        missing.append("pandas")
    raise ImportError(
        "This function requires scikit-learn and "
        f"pandas to be installed. Missing {','.join(missing)}."
    )


__all__ = [
    "path_silhouette",
    "boundary_silhouette",
    "silhouette_alist",
    "nearest_label",
]


def path_silhouette(
    data,
    labels,
    W,
    D=None,
    metric=skp.euclidean_distances,
    closest=False,
    return_nbfc=False,
    return_nbfc_score=False,
    return_paths=False,
    directed=False,
):
    """
    Compute a path silhouette for all observations
    :cite:`wolf2019geosilhouettes,Rousseeuw1987`.


    Parameters
    ----------
    data    :   np.ndarray (N,P)
                matrix of data with N observations and P covariates.
    labels  :   np.ndarray (N,)
                flat vector of the L labels assigned over N observations.
    W       :   libpysal.weights.W | libpysal.graph.Graph
                spatial weights object reflecting the spatial connectivity
                in the problem under analysis
    D       :   np.ndarray (N,N)
                a precomputed distance matrix to apply over W. If passed,
                takes precedence over data, and data is ignored.
    metric  :   callable
                function mapping the (N,P) data into an (N,N) dissimilarity matrix,
                like that found in scikit.metrics.pairwise or scipy.spatial.distance
    closest :   bool
                whether or not to consider the observation "connected" when it
                is first connected to the cluster, or considering the path cost
                to transit through the cluster. If True, the path cost is assessed
                between i and the path-closest j in each cluster. If False, the path
                cost is assessed as the average of path costs between i and all j
                in each cluster
    return_nbfc     :   bool
                        Whether or not to return the label of the next best fit
                        cluster
    return_nbfc_score:  bool
                        Whether or not to return the score of the next best fit
                        cluster.
    return_paths    :   bool
                        Whether or not to return the matrix of shortest path
                        lengths after having computed them.
    directed    :   bool
                    whether to consider the weights matrix as directed or undirected.
                    If directed, asymmetry in the input W is heeded. If not,
                    asymmetry is ignored.

    Returns
    -------
    An (N_obs,) array of the path silhouette values for each observation.
    """
    if not HAS_REQUIREMENTS:
        _raise_initial_error()

    if D is None:
        D = metric(data)
    # polymorphic for sparse & dense input
    assert (D < 0).sum() == 0, (
        "Distance metric has negative values, which is not supported."
    )
    off_diag_zeros = (D + np.eye(D.shape[0])) == 0
    D[off_diag_zeros] = -1
    Wm = sp.csr_matrix(W.sparse)
    DW = sp.csr_matrix(Wm.multiply(D))
    DW.eliminate_zeros()
    DW[DW < 0] = 0
    assert (DW < 0).sum() == 0
    all_pairs = cg.shortest_path(DW, directed=directed)
    labels = np.asarray(labels)
    if W.n_components > 1:
        from libpysal.weights.util import WSP

        psils_ = np.empty(W.n, dtype=float)
        closest_connecting_label_ = np.empty(W.n, dtype=labels.dtype)
        closest_connection_score_ = np.empty(W.n, dtype=labels.dtype)
        for component in np.unique(W.component_labels):
            this_component_mask = np.nonzero(W.component_labels == component)[0]
            subgraph = W.sparse[
                this_component_mask.reshape(-1, 1),  # these rows
                this_component_mask.reshape(1, -1),
            ]  # these columns
            subgraph_W = WSP(subgraph).to_W()
            assert subgraph_W.n_components == 1
            # DW operation is idempotent
            subgraph_D = DW[
                this_component_mask.reshape(-1, 1),  # these rows
                this_component_mask.reshape(1, -1),
            ]  # these columns
            subgraph_labels = labels[this_component_mask]
            n_subgraph_labels = len(np.unique(subgraph_labels))
            if not (2 < n_subgraph_labels < (subgraph_W.n - 1)):
                psils = subgraph_solutions = [0] * subgraph_W.n
                closest_connecting_label = [np.nan] * subgraph_W.n
                closest_connection_score = [np.inf] * subgraph_W.n
            else:
                subgraph_solutions = path_silhouette(
                    data=None,
                    labels=subgraph_labels,
                    W=subgraph_W,
                    D=subgraph_D,
                    metric=metric,
                    closest=closest,
                    return_nbfc=return_nbfc,
                    return_nbfc_score=return_nbfc_score,
                    return_paths=return_paths,
                    directed=directed,
                )
                # always throw away all_pairs, since we already have it built
                if (return_nbfc or return_nbfc_score) and return_paths:
                    if return_nbfc_score:
                        (
                            psils,
                            closest_connecting_label,
                            closest_connection_score,
                            _,
                        ) = subgraph_solutions
                    else:
                        psils, closest_connecting_label, _ = subgraph_solutions
                elif return_nbfc_score:
                    (
                        psils,
                        closest_connecting_label,
                        closest_connection_score,
                    ) = subgraph_solutions
                elif return_nbfc:
                    psils, closest_connecting_label = subgraph_solutions
                elif return_paths:
                    psils, _ = subgraph_solutions
                else:
                    psils = subgraph_solutions
            if return_nbfc:
                closest_connecting_label_[this_component_mask] = (
                    closest_connecting_label
                )
            if return_nbfc_score:
                closest_connection_score_[this_component_mask] = (
                    closest_connection_score
                )
            psils_[this_component_mask] = psils
        closest_connection_score = closest_connection_score_
        closest_connecting_label = closest_connecting_label_
        psils = psils_
    # Single Connected Component
    elif closest is False:
        psils = sk.silhouette_samples(all_pairs, labels, metric="precomputed")
        if return_nbfc or return_nbfc_score:
            closest_connecting_label = []
            closest_connection_score = []
            for i, label in enumerate(labels):
                row = all_pairs[i].copy()
                in_label = labels == label
                masked_label = row.copy()  # for observations in the row
                masked_label[in_label] = np.inf  # make those in cluster infinite
                nearest_not_in_cluster = np.argmin(masked_label)  # find the closest
                nearest_not_in_cluster_label = labels[nearest_not_in_cluster]  # label
                nearest_not_in_cluster_score = masked_label[nearest_not_in_cluster]
                closest_connecting_label.append(nearest_not_in_cluster_label)
                closest_connection_score.append(nearest_not_in_cluster_score)
    else:
        psils = []
        closest_connecting_label = []
        closest_connection_score = []
        for i, label in enumerate(labels):
            row = all_pairs[i]
            in_label = labels == label
            # required to make argmin pertain to N, not N - len(in_label)
            masked_label = row.copy()
            masked_label[in_label] = np.inf
            nearest_not_in_cluster = np.argmin(masked_label)
            nearest_not_in_cluster_score = row[nearest_not_in_cluster]
            nearest_not_in_cluster_label = labels[nearest_not_in_cluster]

            average_interconnect_in_cluster = row[in_label].mean()
            psil = nearest_not_in_cluster_score - average_interconnect_in_cluster
            psil /= np.maximum(
                nearest_not_in_cluster_score, average_interconnect_in_cluster
            )
            psils.append(psil)
            closest_connecting_label.append(nearest_not_in_cluster_label)
            closest_connection_score.append(nearest_not_in_cluster_score)
        psils = np.asarray(psils)
    if (return_nbfc or return_nbfc_score) and return_paths:
        if return_nbfc_score:
            out = (
                psils,
                np.asarray(closest_connecting_label),
                np.asarray(closest_connection_score),
                all_pairs,
            )
        else:
            out = psils, np.asarray(closest_connecting_label), all_pairs
    elif return_nbfc_score:
        out = (
            psils,
            np.asarray(closest_connecting_label),
            np.asarray(closest_connection_score),
        )
    elif return_nbfc:
        out = psils, np.asarray(closest_connecting_label)
    elif return_paths:
        out = psils, all_pairs
    else:
        out = psils
    return out


def boundary_silhouette(
    data, labels, W, metric=skp.euclidean_distances, drop_islands=True
):
    """
    Compute the observation-level boundary silhouette
    score :cite:`wolf2019geosilhouettes`.

    Parameters
    ----------
    data    :   (N_obs,P) numpy array
                an array of covariates to analyze. Each row should be one
                observation, and each clumn should be one feature.
    labels  :   (N_obs,) array of labels
                the labels corresponding to the group each observation is assigned.
    W       :   libpysal.weights.W | libpysal.graph.Graph
                a spatial weights object containing the connectivity structure
                for the data
    metric  :   callable, array,
                a function that takes an argument (data) and returns the all-pairs
                distances/dissimilarity between observations.
    drop_islands : bool (default True)
        Whether or not to preserve islands as entries in the adjacency
        list. By default, observations with no neighbors do not appear
        in the adjacency list. If islands are kept, they are coded as
        self-neighbors with zero weight. See ``libpysal.weights.to_adjlist()``.

    Returns
    -------
    (N_obs,) array of boundary silhouette values for each observation

    Notes
    -----
    The boundary silhouette is the silhouette score using only spatially-proximate
    clusters as candidates for the next-best-fit distance function (the
    b(i) function in :cite:`Rousseeuw1987`.
    This restricts the next-best-fit cluster to be the set of clusters on which
    an observation neighbors.
    So, instead of considering *all* clusters when finding the next-best-fit cluster,
    only clusters that `i` borders are considered.
    This is supposed to model the fact that, in spatially-constrained clustering,
    observation i can only be reassigned from cluster c to cluster k if
    some observation j neighbors i and also resides in k.

    If an observation only neighbors its own cluster, i.e. is not on the boundary
     of a cluster, this value is zero.

    If a cluster has exactly one observation, this value is zero.

    If an observation is on the boundary of more than one cluster, then the
    best candidate is chosen from the set of clusters on which the observation borders.

    metric is a callable mapping an (N,P) data into an (N,N) distance matrix OR
    an (N,N) distance matrix already.
    """
    if not HAS_REQUIREMENTS:
        _raise_initial_error()

    if isinstance(W, weights.W):
        alist = W.to_adjlist(drop_islands=drop_islands)
        index = W.id_order
    else:
        if drop_islands:
            alist = W.adjacency.drop(W.isolates).reset_index()
        else:
            alist = W.adjacancy.reset_index()
        index = W.unique_ids
    labels = np.asarray(labels)
    if callable(metric):
        full_distances = metric(data)
    elif isinstance(metric, np.ndarray):
        n_obs = W.n
        if metric.shape == (n_obs, n_obs):
            full_distances = metric
        else:
            raise ValueError(
                "Precomputed metric is supplied, but is not the right shape."
                f" The dissimilarity matrix should be of shape ({W.n},{W.n}), "
                f" but was of shape ({metric.shape})."
            )
    else:
        raise ValueError(
            "The provided metric is neither a dissmilarity function"
            " nor a dissimilarity matrix."
        )
    assert (full_distances < 0).sum() == 0, (
        "Distance metric has negative values, which is not supported"
    )
    label_frame = pd.DataFrame(labels, index=index, columns=["label"])
    alist = alist.merge(
        label_frame, left_on="focal", right_index=True, how="left"
    ).merge(
        label_frame,
        left_on="neighbor",
        right_index=True,
        how="left",
        suffixes=("_focal", "_neighbor"),
    )
    alist["boundary"] = alist.label_focal != alist.label_neighbor
    focals = alist.groupby("focal")
    bmask = focals.boundary.any()
    result = []
    np.seterr(all="raise")
    for i, (ix, bnd) in enumerate(bmask.items()):
        if not bnd:
            result.append(np.array([0]))
            continue
        sil_score = np.array([np.inf])
        label = labels[i]
        focal_mask = np.nonzero(labels == label)[0]
        if len(focal_mask) == 1:  # the candidate is singleton
            result.append(np.array([0]))
            continue
        neighbors = alist.query(f"focal == {ix}").label_neighbor
        mean_dissim = full_distances[i, focal_mask].sum() / (len(focal_mask) - 1)
        if not np.isfinite(mean_dissim).all():
            raise ValueError(
                "A non-finite mean dissimilarity between groups "
                "and the boundary observation occurred. Please ensure "
                "the data & labels are formatted and shaped correctly."
            )
        neighbor_score = np.array([np.inf])
        for neighbor in set(neighbors).difference([label]):
            other_mask = np.nonzero(labels == neighbor)[0]
            other_score = full_distances[i, other_mask].mean()
            neighbor_score = np.minimum(neighbor_score, other_score)
            if neighbor_score < 0:
                raise ValueError(
                    "A negative neighborhood similarity value occurred. This should "
                    "not happen. Please create a bug report on "
                    "https://github.com/pysal/esda/issues"
                )
        sil_score = (neighbor_score - mean_dissim) / np.maximum(
            neighbor_score, mean_dissim
        )
        result.append(sil_score)
    if len(result) != len(labels):
        raise ValueError(
            "The number of boundary silhouettes does not match the number of "
            "observations. This should not happen. Please create a bug report on "
            "https://github.com/pysal/esda/issues"
        )
    return np.asarray(result).squeeze()


def silhouette_alist(data, labels, alist, indices=None, metric=skp.euclidean_distances):
    """
    Compute the silhouette for each edge in an adjacency graph. Given the alist
    containing `focal` id, `neighbor` id, and `label_focal`, and `label_neighbor`,
    this computes:

    .. math::

        d(i,label_neighbor) - d(i,label_focal)
        / (max(d(i,label_neighbor), d(i,label_focal)))

    Parameters
    ----------
    data : (N,P) array to cluster on or DataFrame indexed on the same values as
           that in alist.focal/alist.neighbor
    labels: (N,) array containing classifications, indexed on the same values
                 as that in alist.focal/alist.neighbor
    alist: adjacency list containing columns focal & neighbor,
           describing one edge of the graph.
    indices: (N,) array containing the "name" for observations in
           alist to be linked to data. indices should be:
           1. aligned with data by iteration order
           2. include all values in the alist.focal set.
           if alist.focal and alist.neighbor are strings, then indices should be
           a list/array of strings aligned with the rows of data.
           if not provided and labels is a series/dataframe,
           then its index will be used.
    metric  :   callable, array,
                a function that takes an argument (data) and returns the all-pairs
                distances/dissimilarity between observations.

    Returns
    -------
    pandas.DataFrame, copy of the adjacency list `alist`, with an additional
    column called `silhouette` that contains the pseudo-silhouette values
    expressing the relative dissimilarity between neighboring observations.
    """
    if not HAS_REQUIREMENTS:
        _raise_initial_error()

    n_obs = data.shape[0]
    if callable(metric):
        full_distances = metric(data)
    elif isinstance(metric, np.ndarray) and metric.shape == (n_obs, n_obs):
        full_distances = metric
    if isinstance(data, pd.DataFrame):
        indices = data.index
    if isinstance(labels, pd.DataFrame | pd.Series) and indices is None:
        indices = labels.index
    elif indices is not None and not isinstance(labels, pd.DataFrame | pd.Series):
        labels = pd.Series(labels, index=indices)
    elif indices is None and not isinstance(labels, pd.DataFrame | pd.Series):
        indices = np.arange(len(labels))
        labels = pd.Series(labels, index=indices)
    if isinstance(labels, pd.DataFrame):
        labels = pd.Series(labels.values, index=labels.index)
    assert indices is not None
    assert isinstance(labels, pd.Series)
    labels = labels.to_frame("label")

    result = alist.sort_values("focal").copy(deep=True)

    result = result.merge(labels, left_on="focal", right_index=True, how="left").merge(
        labels,
        left_on="neighbor",
        right_index=True,
        how="left",
        suffixes=("_focal", "_neighbor"),
    )
    self_dcache = dict()
    sils = []
    indices = list(indices)
    for _, row in result.iterrows():
        name = row.focal
        label = row.label_focal
        neighbor_label = row.label_neighbor
        if neighbor_label == label:
            sils.append(0)
            continue
        i_Xc = indices.index(name)
        mask = labels == label
        mask = np.nonzero(mask.values)[0]
        within_cluster = self_dcache.get(
            (i_Xc, label), full_distances[i_Xc, mask].mean()
        )
        self_dcache[(i_Xc, label)] = within_cluster
        neighbor_mask = labels == neighbor_label
        neighbor_mask = np.nonzero(neighbor_mask.values)[0]
        if len(neighbor_mask) == 0:
            sils.append(0)
            warnings.warn(
                f"A link ({row.focal},{row.neighbor}) has been found to have an empty "
                "set of neighbors. This may happen when a label assignment is "
                "missing for the neighbor unit. Check that no labels are missing.",
                UserWarning,
                stacklevel=2,
            )
            continue
        outer_distance = full_distances[i_Xc, neighbor_mask].mean()
        dist_diff = outer_distance - within_cluster
        dist_max = np.maximum(outer_distance, within_cluster)
        sils.append(dist_diff / dist_max)
    result["silhouette"] = sils
    return result.sort_values("focal").reset_index(drop=True)


def nearest_label(
    data, labels, metric=skp.euclidean_distances, return_distance=False, keep_self=False
):
    """
    Find the nearest label in attribute space.

    Given the data and a set of labels in labels, this finds the label
    whose mean center is closest to the observation in data.

    Parameters
    ----------
    data : (N,P) array to cluster on or DataFrame indexed on the same values as
        that in alist.focal/alist.neighbor
    labels : (N,) array containing classifications, indexed on the same values
        as that in alist.focal/alist.neighbor
    metric : callable, array,
        a function that takes an argument (data) and returns the all-pairs
        distances/dissimilarity between observations.
    return_distance: bool
        Whether to return the distance from the observation to its nearest
        cluster in feature space. If True, the tuple of (nearest_label, dissim)
        is returned. If False, only the nearest_label array is returned.
    keep_self:  bool
        whether to allow observations to use their current cluster as their
        nearest label. If True, an observation's existing cluster assignment can
        also be the cluster it is closest to. If False, an observation's existing
        cluster assignment cannot be the cluster it is closest to. This would mean
        the function computes the nearest *alternative* cluster.

    Returns
    -------
    (N_obs,) array of assignments reflect each observation's nearest label.

    If return_distance is True, a tuple of ((N,) and (N,)) where the first
        array is the assignment, and the second is the distance to the centroid
        of that assignment.
    """
    if not HAS_REQUIREMENTS:
        _raise_initial_error()

    if callable(metric):
        dissim = metric(data)
    elif metric.lower == "precomputed":
        assert data.shape == (
            labels.shape[0],
            labels.shape[0],
        ), "Dissimilarity matrix is malformed!"
        dissim = data
    elif isinstance(metric, np.ndarray):
        assert metric.shape == (
            labels.shape[0],
            labels.shape[0],
        ), "Dissimilarity matrix is malformed!"
        dissim = metric
    unique_labels = np.unique(labels)
    nearest_label = np.empty(labels.shape, dtype=labels.dtype)
    nearest_label_dissim = np.empty(labels.shape)
    for label in unique_labels:
        this_label_mask = labels == label
        this_label_mask = np.nonzero(this_label_mask)[0]
        next_best_fit = np.ones(this_label_mask.shape) * np.inf
        next_best_label = np.empty(this_label_mask.shape, dtype=labels.dtype)
        for neighbor in unique_labels:
            if (neighbor == label) & (not keep_self):
                continue
            neighbor_label_mask = labels == neighbor
            n_in_neighbor = neighbor_label_mask.sum()
            neighbor_label_mask = np.nonzero(neighbor_label_mask)[0].reshape(1, -1)
            # Need to account for the fact that the self-distance
            # is not included in the silhouette; in small clusters,
            # this extra zero can bring down the average, resulting in a case
            # where the silhouette is negative, but the "nearest" cluster would
            # be the current cluster if we take averages including i in C.
            chunk = dissim[
                this_label_mask.reshape(-1, 1), neighbor_label_mask  # these rows
            ]  # and these columns
            neighbor_distance = chunk.sum(axis=1) / np.maximum(
                n_in_neighbor - 1, 1
            )  # and sum across rows
            next_best_label[neighbor_distance < next_best_fit] = neighbor
            next_best_fit = np.minimum(next_best_fit, neighbor_distance)
        nearest_label[this_label_mask] = next_best_label
        nearest_label_dissim[this_label_mask] = next_best_fit
    if return_distance:
        return nearest_label, nearest_label_dissim
    else:
        return nearest_label



"""
Geary's C statistic for spatial autocorrelation
"""

__author__ = "Serge Rey <sjsrey@gmail.com> "

import warnings

import numpy as np
import scipy.stats as stats
from libpysal import graph, weights

from .tabular import _univariate_handler

__all__ = ["Geary"]


class Geary:
    """
    Global Geary C Autocorrelation statistic

    Parameters
    ----------
    y              : array
                     (n, 1) attribute vector
    w              : W | Graph
                     spatial weights instance as W or Graph aligned with y
    transformation : {'R', 'B', 'D', 'U', 'V'}
                     weights transformation, default is row-standardized.
                     Other options include "B": binary, "D":
                     doubly-standardized, "U": untransformed (general
                     weights), "V": variance-stabilizing.
    permutations   : int
                     number of random permutations for calculation of
                     pseudo-p_values

    Attributes
    ----------
    y              : array
                     original variable
    w              : W | Graph
                     spatial weights
    permutations   : int
                     number of permutations
    C              : float
                     value of statistic
    EC             : float
                     expected value
    VC             : float
                     variance of G under normality assumption
    z_norm         : float
                     z-statistic for C under normality assumption
    z_rand         : float
                     z-statistic for C under randomization assumption
    p_norm         : float
                     p-value under normality assumption (one-tailed)
    p_rand         : float
                     p-value under randomization assumption (one-tailed)
    sim            : array
                     (if permutations!=0)
                     vector of I values for permutated samples
    p_sim          : float
                     (if permutations!=0)
                     p-value based on permutations (one-tailed)
                     null: sptial randomness
                     alternative: the observed C is extreme
                     it is either extremely high or extremely low
    EC_sim         : float
                     (if permutations!=0)
                     average value of C from permutations
    VC_sim         : float
                     (if permutations!=0)
                     variance of C from permutations
    seC_sim        : float
                     (if permutations!=0)
                     standard deviation of C under permutations.
    z_sim          : float
                     (if permutations!=0)
                     standardized C based on permutations
    p_z_sim        : float
                     (if permutations!=0)
                     p-value based on standard normal approximation from
                     permutations (one-tailed)

    Examples
    --------
    >>> import libpysal
    >>> from esda.geary import Geary
    >>> w = libpysal.io.open(libpysal.examples.get_path("book.gal")).read()
    >>> f = libpysal.io.open(libpysal.examples.get_path("book.txt"))
    >>> y = np.array(f.by_col['y'])
    >>> c = Geary(y,w,permutations=0)
    >>> round(c.C,7)
    0.3330108
    >>> round(c.p_norm,7)
    9.2e-05
    >>>


    Notes
    -----
    Technical details and derivations can be found in :cite:`cliff81`.

    """

    def __init__(self, y, w, transformation="r", permutations=999):
        if not isinstance(w, weights.W | graph.Graph):
            raise TypeError(
                "w must be a libpysal.weights.W or libpysal.graph.Graph object, "
                f"got {type(w)} instead."
            )
        y = np.asarray(y).flatten()
        self.n = len(y)
        self.y = y
        if isinstance(w, weights.W):
            w.transform = transformation
        else:
            w = w.transform(transformation)
            self.summary = w.summary()
        self.w = w
        self._focal_ix, self._neighbor_ix = w.sparse.nonzero()
        self._weights = w.sparse.data
        self.permutations = permutations
        self.__moments()
        xn = range(len(y))
        self.xn = xn
        self.y2 = y * y
        yd = y - y.mean()
        yss = sum(yd * yd)
        s0 = self.w.s0 if isinstance(self.w, weights.W) else self.summary.s0

        self.den = yss * s0 * 2.0
        self.C = self.__calc(y)
        de = self.C - 1.0
        self.EC = 1.0
        self.z_norm = de / self.seC_norm
        self.z_rand = de / self.seC_rand
        if de > 0:
            self.p_norm = stats.norm.sf(self.z_norm)
            self.p_rand = stats.norm.sf(self.z_rand)
        else:
            self.p_norm = stats.norm.cdf(self.z_norm)
            self.p_rand = stats.norm.cdf(self.z_rand)

        if permutations:
            sim = [
                self.__calc(np.random.permutation(self.y)) for i in range(permutations)
            ]
            self.sim = sim = np.array(sim)
            above = sim >= self.C
            larger = sum(above)
            if (permutations - larger) < larger:
                larger = permutations - larger
            self.p_sim = (larger + 1.0) / (permutations + 1.0)
            self.EC_sim = sum(sim) / permutations
            self.seC_sim = np.array(sim).std()
            self.VC_sim = self.seC_sim**2
            self.z_sim = (self.C - self.EC_sim) / self.seC_sim
            self.p_z_sim = stats.norm.sf(np.abs(self.z_sim))

    @property
    def _statistic(self):
        """a standardized accessor for esda statistics"""
        return self.C

    def __moments(self):
        y = self.y
        n = self.n
        s0 = self.w.s0 if isinstance(self.w, weights.W) else self.summary.s0
        s1 = self.w.s1 if isinstance(self.w, weights.W) else self.summary.s1
        s2 = self.w.s2 if isinstance(self.w, weights.W) else self.summary.s2
        s02 = s0 * s0
        yd = y - y.mean()
        yd4 = yd**4
        yd2 = yd**2
        n2 = n * n
        k = (yd4.sum() / n) / ((yd2.sum() / n) ** 2)
        A = (n - 1) * s1 * (n2 - 3 * n + 3 - (n - 1) * k)
        B = (1.0 / 4) * ((n - 1) * s2 * (n2 + 3 * n - 6 - (n2 - n + 2) * k))
        C = s02 * (n2 - 3 - (n - 1) ** 2 * k)
        vc_rand = (A - B + C) / (n * (n - 2) * (n - 3) * s02)
        vc_norm = (1 / (2 * (n + 1) * s02)) * ((2 * s1 + s2) * (n - 1) - 4 * s02)

        self.VC_rand = vc_rand
        self.VC_norm = vc_norm
        self.seC_rand = vc_rand ** (0.5)
        self.seC_norm = vc_norm ** (0.5)

    def __calc(self, y):
        num = (self._weights * ((y[self._focal_ix] - y[self._neighbor_ix]) ** 2)).sum()
        a = (self.n - 1) * num
        return a / self.den

    @classmethod
    def by_col(
        cls, df, cols, w=None, inplace=False, pvalue="sim", outvals=None, **stat_kws
    ):
        """
        Function to compute a Geary statistic on a dataframe

        Parameters
        ----------
        df : pandas.DataFrame
            a pandas dataframe with a geometry column
        cols : string or list of string
            name or list of names of columns to use to compute the statistic
        w : pysal weights object
            a weights object aligned with the dataframe. If not provided, this
            is searched for in the dataframe's metadata
        inplace : bool
            a boolean denoting whether to operate on the dataframe inplace or to
            return a series contaning the results of the computation. If
            operating inplace, with default configurations,
            the derived columns will be named like 'column_geary' and 'column_p_sim'
        pvalue  : string
            a string denoting which pvalue should be returned. Refer to the
            the Geary statistic's documentation for available p-values
        outvals : list of strings
            list of arbitrary attributes to return as columns from the
            Geary statistic
        **stat_kws : dict
            options to pass to the underlying statistic. For this, see the
            documentation for the Geary statistic.

        Returns
        --------
        If inplace, None, and operation is conducted on dataframe in memory. Otherwise,
        returns a copy of the dataframe with the relevant columns attached.

        Notes
        -----
        Technical details and derivations can be found in :cite:`cliff81`.

        """

        msg = (
            "The `.by_col()` methods are deprecated and will be "
            "removed in a future version of `esda`."
        )
        warnings.warn(msg, FutureWarning, stacklevel=2)

        return _univariate_handler(
            df,
            cols,
            w=w,
            inplace=inplace,
            pvalue=pvalue,
            outvals=outvals,
            stat=cls,
            swapname=cls.__name__.lower(),
            **stat_kws,
        )



import pathlib

import geopandas
import numpy
import pytest

from .. import map_comparison as mc

shapely = pytest.importorskip("shapely")

filepath = pathlib.Path(__file__).parent / "regions.zip"

r1 = geopandas.read_file(filepath, layer="regions1")
r1a = mc._cast(r1.geometry)
r2 = geopandas.read_file(filepath, layer="regions2")
r2a = mc._cast(r2.geometry)


def test_areal_entropy():
    ae1 = mc.areal_entropy(r1, base=2)
    ae1e = mc.areal_entropy(r1)
    numpy.testing.assert_equal(2, ae1)
    numpy.testing.assert_allclose(ae1e, 1.38629436111)
    ae1l = mc.areal_entropy(r1, base=2, local=True)
    numpy.testing.assert_allclose(ae1l, numpy.ones((4,)) * 0.5)
    ae1_from_areas = mc.areal_entropy(areas=shapely.area(r1a), base=2)
    numpy.testing.assert_equal(ae1, ae1_from_areas)


def test_overlay_entropy():
    oe = mc.overlay_entropy(r1, r2)
    oe2 = mc.overlay_entropy(r1, r2, base=2)
    # standardization removes base sensitivity
    numpy.testing.assert_equal(oe, oe2)
    oe2_raw = mc.overlay_entropy(r1, r2, base=2, standardize=False)
    oe_raw = mc.overlay_entropy(r1, r2, standardize=False)
    with pytest.raises(AssertionError):
        numpy.testing.assert_equal(oe2_raw, oe_raw)
        numpy.testing.assert_equal(oe2_raw, oe2)
    oe2_local = mc.overlay_entropy(r1, r2, base=2, local=True)
    oe2_local_raw = mc.overlay_entropy(r1, r2, base=2, local=True, standardize=False)
    numpy.testing.assert_equal(oe2_local.shape, (4,))
    numpy.testing.assert_equal(oe2_local.sum(), oe2)
    numpy.testing.assert_equal(oe2_local_raw.sum(), oe2_raw)


def test_completeness_and_homogeneity():
    for base in (numpy.e, 2):
        c1 = mc.completeness(r1, r2, base=base, local=True)
        c2 = mc.completeness(r2, r1, base=base, local=True)
        h1 = mc.homogeneity(r1, r2, base=base, local=True)
        h2 = mc.homogeneity(r2, r1, base=base, local=True)

        numpy.testing.assert_array_equal(c1, h2)
        numpy.testing.assert_array_equal(c2, h1)
        numpy.testing.assert_array_equal(c1.sum(), mc.completeness(r1, r2, base=base))

        numpy.testing.assert_equal(c1.shape, r1.geometry.shape)
        numpy.testing.assert_equal(c2.shape, r2.geometry.shape)
        numpy.testing.assert_allclose(c1.sum(), 0.42275150844)
        numpy.testing.assert_allclose(h1.sum(), 0.31534179219)


def test_external_entropy():
    for base in (2, numpy.e):
        v1 = mc.external_entropy(r1, r2, base=base)
        v2 = mc.external_entropy(r2, r1, base=base)
        numpy.testing.assert_allclose(v1, v2)
        numpy.testing.assert_allclose(v1, 0.3612313462)
        v1_ctipped = mc.external_entropy(r1, r2, base=base, balance=100)
        numpy.testing.assert_allclose(
            v1_ctipped, 0.42275150844
        )  # value from completeness
        v1_htipped = mc.external_entropy(r1, r2, base=base, balance=-100)
        numpy.testing.assert_allclose(
            v1_htipped, 0.31534179219
        )  # value from homogeneity



"""
Centralised conditional randomisation engine. Numba accelerated.
"""

import os
import warnings
from .significance import _permutation_significance
import numpy as np

try:
    from numba import boolean, njit, prange
except (ImportError, ModuleNotFoundError):
    from libpysal.common import jit as njit
    prange = range
    boolean = bool


__all__ = ["crand"]

#######################################################################
#                   Utilities for all functions                       #
#######################################################################


@njit(fastmath=True)
def vec_permutations(
    max_card: int, n: int, k_replications: int, seed: int
    ):
    """
    Generate `max_card` permuted IDs, sampled from `n` without replacement,
    `k_replications` times
    ...

    Parameters
    ----------
    max_card : int
        Number of permuted IDs to generate per sample
    n : int
        Size of the sample to sample IDs from
    k_replications : int
        Number of samples of permuted IDs to perform
    seed : int
        Seed to ensure reproducibility of conditional randomizations

    Returns
    -------
    result : ndarray
        (k_replications, max_card) array with permuted IDs
    """
    np.random.seed(seed)
    result = np.empty((k_replications, max_card), dtype=np.int64)
    for k in prange(k_replications):
        result[k] = np.random.choice(n - 1, size=max_card, replace=False)
    return result


def crand(
    z,
    w,
    observed,
    permutations,
    keep,
    n_jobs,
    stat_func,
    scaling=None,
    seed=None,
    island_weight=0,
    alternative=None
):
    """
    Conduct conditional randomization of a given input using the provided
    statistic function. Numba accelerated.
    ...

    Parameters
    ----------
    z : ndarray
        2D array with N rows with standardised observed values
    w : libpysal.weights.W
        Spatial weights object
    observed : ndarray
        (N,) array with observed values
    permutations : int
        Number of permutations for conditional randomisation
    keep : Boolean
        If True, store simulation; else do not return randomised statistics
    n_jobs : int
        Number of cores to be used in the conditional randomisation. If -1,
        all available cores are used.
    stat_func : callable
        Method implementing the spatial statistic to be evaluated under
        conditional randomisation. The method needs to have the following
        signature:
            i : int
                Position of observation to be evaluated in the sample
            z : ndarray
                2D array with N rows with standardised observed values
            permuted_ids : ndarray
                (permutations, max_cardinality) array with indices of permuted
                IDs
            weights_i : ndarray
                Weights for neighbors in i
            scaling : float
                Scaling value to apply to every local statistic
     seed : None/int
        Seed to ensure reproducibility of conditional randomizations

    Returns
    -------
    p_sim : ndarray
        (N,) array with pseudo p-values from conditional permutation
    rlocals : ndarray
        If keep=True, (N, permutations) array with simulated values
        of stat_func under the null of spatial randomness; else, empty (1, 1) array
    """
    adj_matrix = w.sparse

    n = len(z)
    if z.ndim == 2:
        if z.shape[1] == 2:
            # assume that matrix is [X Y], and scaling is moran-like
            scaling = (
                (n - 1) / (z[:, 0] * z[:, 0]).sum() if (scaling is None) else scaling
            )
        elif z.shape[1] == 1:
            # assume that matrix is [X], and scaling is moran-like
            scaling = (n - 1) / (z * z).sum() if (scaling is None) else scaling
        else:
            raise NotImplementedError(
                f"multivariable input is not yet supported in "
                f"conditional randomization. Recieved `z` of shape {z.shape}"
            )
    elif z.ndim == 1:
        scaling = (n - 1) / (z * z).sum() if (scaling is None) else scaling
    else:
        raise NotImplementedError(
            f"multivariable input is not yet supported in "
            f"conditional randomization. Recieved `z` of shape {z.shape}"
        )

    if alternative is None:
        warnings.warn(
            "The alternative hypothesis for conditional randomization"
            " is changing in the next major release of esda. We recommend"
            " setting alternative='two-sided', which will generally"
            " double the p-value returned." 
            " To retain the current behavior, set alternative='directed'."
            " We strongly recommend moving to alternative='two-sided'.",
            DeprecationWarning,
        )
        # TODO: replace this with 'two-sided' by next major release
        alternative = 'directed'
    if alternative not in ("two-sided", "greater", "lesser", "directed", "folded"):
        raise ValueError(
            f"alternative='{alternative}' provided, but is not"
            f" one of the supported options: 'two-sided', 'greater', 'lesser', 'directed', 'folded')"
        )

    # paralellise over permutations?
    if seed is None:
        seed = np.random.randint(12345, 12345000)

    # we need to be careful to shuffle only *other* sites, not
    # the self-site. This means we need to
    # extract the self-weight, if any
    self_weights = adj_matrix.diagonal()
    # force the self-site weight to zero
    with warnings.catch_warnings():
        # massive changes to sparsity incur a cost, but it's not
        # large for simply changing the diag
        warnings.simplefilter("ignore")
        adj_matrix.setdiag(0)
        adj_matrix.eliminate_zeros()
    # extract the weights from a now no-self-weighted adj_matrix
    other_weights = adj_matrix.data.astype(z.dtype) # cast is forced by @ in numba
    # use the non-self weight as the cardinality, since
    # this is the set we have to randomize.
    # if there is a self-neighbor, we need to *not* shuffle the
    # self neighbor, since conditional randomization conditions on site i.
    cardinalities = np.array((adj_matrix != 0).sum(1)).flatten()
    max_card = cardinalities.max()
    permuted_ids = vec_permutations(max_card, n, permutations, seed)

    if n_jobs != 1:
        try:
            import joblib  # noqa: F401
        except (ModuleNotFoundError, ImportError):
            warnings.warn(
                f"Parallel processing is requested (n_jobs={n_jobs}),"
                f" but joblib cannot be imported. n_jobs will be set"
                f" to 1.",
                stacklevel=2,
            )
            n_jobs = 1

    if n_jobs == 1:
        p_sims, rlocals = compute_chunk(
            0,  # chunk start
            z,  # chunked z, for serial this is the entire data
            z,  # all z, for serial this is also the entire data
            observed,  # observed statistics
            cardinalities,  # cardinalities conforming to chunked z
            self_weights,  # n-length vector containing the self-weights.
            other_weights,  # flat weights buffer
            permuted_ids,  # permuted ids
            scaling,  # scaling applied to all statistics
            keep,  # whether or not to keep the local statistics
            stat_func,
            island_weight,
            alternative=alternative
        )
    else:
        if n_jobs == -1:
            n_jobs = os.cpu_count()
        if n_jobs > len(z):
            n_jobs = len(z)
        # Parallel implementation
        p_sims, rlocals = parallel_crand(
            z,
            observed,
            cardinalities,
            self_weights,
            other_weights,
            permuted_ids,
            scaling,
            n_jobs,
            keep,
            stat_func,
            island_weight,
            alternative=alternative
        )

    return p_sims, rlocals


@njit(parallel=False, fastmath=True)
def compute_chunk(
    chunk_start: int,
    z_chunk: np.ndarray,
    z: np.ndarray,
    observed: np.ndarray,
    cardinalities: np.ndarray,
    self_weights: np.ndarray,
    other_weights: np.ndarray,
    permuted_ids: np.ndarray,
    scaling: np.float64,
    keep: bool,
    stat_func,
    island_weight: float,
    alternative: str 
):
    """
    Compute conditional randomisation for a single chunk
    ...

    Parameters
    ----------
    chunk_start : int
        Starting index for the chunk of input. Should be zero if z_chunk == z.
    z_chunk : numpy.ndarray
        (n_chunk,) array containing the chunk of standardised observed values.
    z : ndarray
        2D array with N rows with standardised observed values
    observed : ndarray
        (n_chunk,) array containing observed values for the chunk
    cardinalities : ndarray
        (n_chunk,) array containing the cardinalities for each element.
    self_weights : ndarray of shape (n,)
        Array containing the self-weights for each observation. In most cases, this
        will be zero. But, in some cases (e.g. Gi-star or kernel weights), this will
        be nonzero.
    other_weights : ndarray
        Array containing the weights of all other sites in the computation
        other than site i. If self_weights is zero, this has as many entries
        as the sum of `cardinalities`.
    permuted_ids : ndarray
        (permutations, max_cardinality) array with indices of permuted
        ids to use to construct random realizations of the statistic
    scaling : float
        Scaling value to apply to every local statistic
    keep : bool
        If True, store simulation; else do not return randomised statistics
    stat_func : callable
        Method implementing the spatial statistic to be evaluated under
        conditional randomisation. The method needs to have the following
        signature:
            i : int
                Position of observation to be evaluated in the sample
            z : ndarray
                2D array with N rows with standardised observed values
            permuted_ids : ndarray
                (permutations, max_cardinality) array with indices of permuted
                IDs
            weights_i : ndarray
                Weights for neighbors in i
            scaling : float
                Scaling value to apply to every local statistic
    island_weight:
        value to use as a weight for the "fake" neighbor for every island.
        If numpy.nan, will propagate to the final local statistic depending
        on the `stat_func`. If 0, then the lag is always zero for islands.

    Returns
    -------
    larger : ndarray
        (n_chunk,) array with number of random draws under the null larger
        than observed value of statistic
    rlocals : ndarray
        (n_chunk, max_cardinality) array with local statistics simulated under
        the null of spatial randomness
    """
    chunk_n = z_chunk.shape[0]
    n_samples = z.shape[0]
    p_permutations, k_max_card = permuted_ids.shape
    p_sims = np.zeros((chunk_n,), dtype=np.float32)
    rlocals = np.empty((chunk_n, permuted_ids.shape[0])) if keep else np.empty((1, 1))


    mask = np.ones((n_samples,), dtype=np.int8) == 1
    wloc = 0

    for i in range(chunk_n):
        cardinality = cardinalities[i]
        if cardinality == 0:  # deal with islands
            weights_i = np.zeros(2, dtype=other_weights.dtype)
            weights_i[1] = island_weight
        else:
            # we need to fix the self-weight to the first position
            weights_i = np.zeros(cardinality + 1, dtype=other_weights.dtype)
            weights_i[0] = self_weights[i]
            # this chomps the next `cardinality` weights off of `weights`
            weights_i[1:] = other_weights[wloc : (wloc + cardinality)]
        wloc += cardinality
        mask[chunk_start + i] = False
        rstats = stat_func(chunk_start + i, z, permuted_ids, weights_i, scaling)
        p_sims[i] = _permutation_significance(
            observed[i], rstats, alternative=alternative
        ).item()
        if keep:
            rlocals[i] = rstats

    return p_sims, rlocals


#######################################################################
#                   Parallel Implementation                           #
#######################################################################


@njit(fastmath=True)
def build_weights_offsets(cardinalities: np.ndarray, n_chunks: int):
    """
    Utility function to construct offsets into the weights
    flat data array found in the W.sparse.data object
    ...

    Parameters
    ----------
    cardinalities : ndarray
        (n_chunk,) array containing the cardinalities for each element.
    n_chunks : int
        Number of chunks to split the weights into

    Returns
    -------
    boundary_points : ndarray
        (n_chunks,) array with positions to split a flat representation of W
        for every chunk
    """
    boundary_points = np.zeros((n_chunks + 1,), dtype=np.int64)
    n = cardinalities.shape[0]
    chunk_size = np.int64(n / n_chunks) + 1
    start = 0
    for i in range(n_chunks):
        advance = cardinalities[start : start + chunk_size].sum()
        boundary_points[i + 1] = boundary_points[i] + advance
        start += chunk_size
    return boundary_points


@njit(fastmath=True)
def chunk_generator(
    n_jobs: int,
    starts: np.ndarray,
    z: np.ndarray,
    observed: np.ndarray,
    cardinalities: np.ndarray,
    self_weights: np.ndarray,
    other_weights: np.ndarray,
    w_boundary_points: np.ndarray,
):
    """
    Construct chunks to iterate over within numba in parallel
    ...

    Parameters
    ----------
    n_jobs : int
        Number of cores to be used in the conditional randomisation. If -1,
        all available cores are used.
    starts : ndarray
        (n_chunks+1,) array of positional starts for each chunk
    z : ndarray
        2D array with N rows with standardised observed values
    observed : ndarray
        (N,) array with observed values
    cardinalities : ndarray
        (N,) array containing the cardinalities for each element.
    weights : ndarray
        Array containing the weights within the chunk in a flat format (ie. as
        obtained from the `values` attribute of a CSR sparse representation of
        the original W. This is as long as the sum of `cardinalities`
    w_boundary_points : ndarray
        (n_chunks,) array with positions to split a flat representation of W
        for every chunk

    Yields
    ------
    start : int
        Starting index for the chunk of input. Should be zero if z_chunk == z.
    z_chunk : numpy.ndarray
        (n_chunk,) array containing the chunk of standardised observed values.
    z : ndarray
        2D array with N rows with standardised observed values
    observed_chunk : ndarray
        (n_chunk,) array containing observed values for the chunk
    cardinalities_chunk : ndarray
        (n_chunk,) array containing the cardinalities for each element.
    weights_chunk : ndarray
        Array containing the weights within the chunk in a flat format (ie. as
        obtained from the `values` attribute of a CSR sparse representation of
        the original W. This is as long as the sum of `cardinalities`
    """
    chunk_size = starts[1] - starts[0]
    for i in range(n_jobs):
        start = starts[i]
        z_chunk = z[start : (start + chunk_size)]
        self_weights_chunk = self_weights[start : (start + chunk_size)]
        observed_chunk = observed[start : (start + chunk_size)]
        cardinalities_chunk = cardinalities[start : (start + chunk_size)]
        w_chunk = other_weights[w_boundary_points[i] : w_boundary_points[i + 1]]
        yield (
            start,
            z_chunk,
            z,
            observed_chunk,
            cardinalities_chunk,
            self_weights_chunk,
            w_chunk,
        )


def parallel_crand(
    z: np.ndarray,
    observed: np.ndarray,
    cardinalities: np.ndarray,
    self_weights: np.ndarray,
    other_weights: np.ndarray,
    permuted_ids: np.ndarray,
    scaling: np.float64,
    n_jobs: int,
    keep: bool,
    stat_func,
    island_weight,
    alternative: str = 'directed'
):
    """
    Conduct conditional randomization in parallel using numba
    ...

    Parameters
    ----------
    z : ndarray
        2D array with N rows with standardised observed values
    observed : ndarray
        (N,) array with observed values
    cardinalities : ndarray
        (N,) array containing the cardinalities for each element.
    self_weights : ndarray of shape (n,)
        Array containing the self-weights for each observation. In most cases, this
        will be zero. But, in some cases (e.g. Gi-star or kernel weights), this will
        be nonzero.
    other_weights : ndarray
        Array containing the weights of all other sites in the computation
        other than site i. If self_weights is zero, this has as many entries
        as the sum of `cardinalities`.
    permuted_ids : ndarray
        (permutations, max_cardinality) array with indices of permuted
        ids to use to construct random realizations of the statistic
    scaling : float64
        Scaling value to apply to every local statistic
    n_jobs : int
        Number of cores to be used in the conditional randomisation. If -1,
        all available cores are used.
    keep : Boolean
        If True, store simulation; else do not return randomised statistics
    stat_func : callable
        Method implementing the spatial statistic to be evaluated under
        conditional randomisation. The method needs to have the following
        signature:
            i : int
                Position of observation to be evaluated in the sample
            z : ndarray
                2D array with N rows with standardised observed values
            permuted_ids : ndarray
                (permutations, max_cardinality) array with indices of permuted
                IDs
            weights_i : ndarray
                Weights for neighbors in i
            scaling : float
                Scaling value to apply to every local statistic
    island_weight:
        value to use as a weight for the "fake" neighbor for every island.
        If numpy.nan, will propagate to the final local statistic depending
        on the `stat_func`. If 0, then the lag is always zero for islands.
    Returns
    -------
    larger : ndarray
        (N,) array with number of random draws under the null larger
        than observed value of statistic
    rlocals : ndarray
        (N, max_cardinality) array with local statistics simulated under
        the null of spatial randomness
    """
    from joblib import Parallel, delayed, parallel_backend

    n = z.shape[0]
    w_boundary_points = build_weights_offsets(cardinalities, n_jobs)
    chunk_size = n // n_jobs + 1
    starts = np.arange(n_jobs + 1) * chunk_size
    # ------------------------------------------------------------------
    # Set up output holders
    rlocals = np.empty((n, permuted_ids.shape[0])) if keep else np.empty((1, 1))

    # ------------------------------------------------------------------
    # Joblib parallel loop by chunks

    # construct chunks using a generator
    chunks = chunk_generator(
        n_jobs,
        starts,
        z,
        observed,
        cardinalities,
        self_weights,
        other_weights,
        w_boundary_points,
    )

    with parallel_backend("loky", inner_max_num_threads=1):
        worker_out = Parallel(n_jobs=n_jobs)(
            delayed(compute_chunk)(
                *pars, permuted_ids, scaling, keep, stat_func, island_weight, alternative
            )
            for pars in chunks
        )

    p_sims, rlocals = zip(*worker_out)
    p_sims = np.hstack(p_sims).squeeze()
    rlocals = np.row_stack(rlocals).squeeze()
    return p_sims, rlocals


#######################################################################
#                   Local statistical functions                       #
#######################################################################


@njit(fastmath=False)
def _prepare_univariate(i, z, permuted_ids, weights_i):
    mask = np.ones_like(z, dtype=boolean)
    mask[i] = False
    z_no_i = z[mask]
    cardinality = len(weights_i)
    flat_permutation_ids = permuted_ids[:, :cardinality].flatten()
    zrand = z_no_i[flat_permutation_ids].reshape(-1, cardinality)
    return z[i], zrand


@njit(fastmath=False)
def _prepare_bivariate(i, z, permuted_ids, weights_i):
    zx = z[:, 0]
    zy = z[:, 1]

    cardinality = len(weights_i)

    mask = np.ones_like(zx, dtype=boolean)
    mask[i] = False
    zx_no_i = zy[mask]
    zy_no_i = zx[mask]

    flat_permutation_indices = permuted_ids[:, :cardinality].flatten()

    zxrand = zx_no_i[flat_permutation_indices].reshape(-1, cardinality)
    zyrand = zy_no_i[flat_permutation_indices].reshape(-1, cardinality)

    return zx[i], zxrand, zy[i], zyrand


@njit(fastmath=True)
def local(i, z, permuted_ids, weights_i, scaling):  # noqa: ARG001
    raise NotImplementedError
    # returns (k_permutations,) array of random statistics for observation i



"""
Getis and Ord G statistic for spatial autocorrelation
"""

__author__ = "Sergio J. Rey <srey@asu.edu>, Myunghwa Hwang <mhwang4@gmail.com> "
__all__ = ["G", "G_Local"]

import warnings

import numpy as np
from libpysal.weights import W
from libpysal.weights.spatial_lag import lag_spatial
from libpysal.weights.util import fill_diagonal
from scipy import stats

from .crand import _prepare_univariate
from .crand import crand as _crand_plus
from .crand import njit as _njit
from .tabular import _univariate_handler

PERMUTATIONS = 999


class G:
    """
    Global G Autocorrelation Statistic

    Parameters
    ----------
    y             : array (n,1)
                    Attribute values
    w             : W | Graph
                   spatial weights instance as W or Graph aligned with y
    permutations  : int
                    the number of random permutations for calculating pseudo p_values

    Attributes
    ----------
    y : array
        original variable
    w : W
        spatial weights instance as W or Graph aligned with y
    permutation : int
        the number of permutations
    G : float
        the value of statistic
    EG : float
        the expected value of statistic
    VG : float
        the variance of G under normality assumption
    z_norm : float
        standard normal test statistic
    p_norm : float
        p-value under normality assumption (one-sided)
    sim : array
        (if permutations > 0)
        vector of G values for permutated samples
    p_sim : float
        p-value based on permutations (one-sided)
        null: spatial randomness
        alternative: the observed G is extreme it is either
        extremely high or extremely low
    EG_sim : float
        average value of G from permutations
    VG_sim : float
        variance of G from permutations
    seG_sim : float
        standard deviation of G under permutations.
    z_sim : float
        standardized G based on permutations
    p_z_sim : float
        p-value based on standard normal approximation from permutations (one-sided)

    Notes
    -----
    Moments are based on normality assumption.

    For technical details see :cite:`Getis_2010` and :cite:`Ord_2010`.


    Examples
    --------
    >>> import libpysal
    >>> import numpy
    >>> numpy.random.seed(10)

    Preparing a point data set

    >>> points = [(10, 10), (20, 10), (40, 10), (15, 20), (30, 20), (30, 30)]

    Creating a weights object from points

    >>> w = libpysal.weights.DistanceBand(points,threshold=15)
    >>> w.transform = "B"

    Preparing a variable

    >>> y = numpy.array([2, 3, 3.2, 5, 8, 7])

    Applying Getis and Ord G test

    >>> from esda.getisord import G
    >>> g = G(y,w)

    Examining the results

    >>> round(g.G, 3)
    0.557

    >>> round(g.p_norm, 3)
    0.173

    """

    def __init__(self, y, w, permutations=PERMUTATIONS):
        y = np.asarray(y).flatten()
        self.n = len(y)
        self.y = y
        if isinstance(w, W):
            w.transform = "B"
        else:
            w = w.transform("B")
            self.summary = w.summary()
        self.w = w
        self.permutations = permutations
        self.__moments()
        self.y2 = y * y
        y = y.reshape(
            len(y), 1
        )  # Ensure that y is an n by 1 vector, otherwise y*y.T == y*y
        self.den_sum = (y * y.T).sum() - (y * y).sum()
        self.G = self.__calc(self.y)
        self.z_norm = (self.G - self.EG) / np.sqrt(self.VG)
        self.p_norm = 1.0 - stats.norm.cdf(np.abs(self.z_norm))

        if permutations:
            sim = [
                self.__calc(np.random.permutation(self.y)) for i in range(permutations)
            ]
            self.sim = sim = np.array(sim)
            above = sim >= self.G
            larger = sum(above)
            if (self.permutations - larger) < larger:
                larger = self.permutations - larger
            self.p_sim = (larger + 1.0) / (permutations + 1.0)
            self.EG_sim = sum(sim) / permutations
            self.seG_sim = sim.std()
            self.VG_sim = self.seG_sim**2
            self.z_sim = (self.G - self.EG_sim) / self.seG_sim
            self.p_z_sim = 1.0 - stats.norm.cdf(np.abs(self.z_sim))

    def __moments(self):
        y = self.y
        n = self.n
        w = self.w
        n2 = n * n
        s0 = w.s0 if isinstance(w, W) else self.summary.s0
        self.EG = s0 / (n * (n - 1))
        s02 = s0 * s0
        s1 = w.s1 if isinstance(w, W) else self.summary.s1
        s2 = w.s2 if isinstance(w, W) else self.summary.s2
        b0 = (n2 - 3 * n + 3) * s1 - n * s2 + 3 * s02
        b1 = (-1.0) * ((n2 - n) * s1 - 2 * n * s2 + 6 * s02)
        b2 = (-1.0) * (2 * n * s1 - (n + 3) * s2 + 6 * s02)
        b3 = 4 * (n - 1) * s1 - 2 * (n + 1) * s2 + 8 * s02
        b4 = s1 - s2 + s02
        self.b0 = b0
        self.b1 = b1
        self.b2 = b2
        self.b3 = b3
        self.b4 = b4
        y2 = y * y
        y3 = y * y2
        y4 = y2 * y2
        EG2 = b0 * (sum(y2) ** 2) + b1 * sum(y4) + b2 * (sum(y) ** 2) * sum(y2)
        EG2 += b3 * sum(y) * sum(y3) + b4 * (sum(y) ** 4)
        EG2NUM = EG2
        EG2DEN = ((sum(y) ** 2 - sum(y2)) ** 2) * n * (n - 1) * (n - 2) * (n - 3)
        self.EG2 = EG2NUM / EG2DEN
        self.VG = self.EG2 - self.EG**2

    def __calc(self, y):
        yl = lag_spatial(self.w, y) if isinstance(self.w, W) else self.w.lag(y)
        self.num = y * yl
        return self.num.sum() / self.den_sum

    @property
    def _statistic(self):
        """Standardized accessor for esda statistics"""
        return self.G

    @classmethod
    def by_col(
        cls, df, cols, w=None, inplace=False, pvalue="sim", outvals=None, **stat_kws
    ):
        """
        Function to compute a G statistic on a dataframe

        Parameters
        ----------
        df : pandas.DataFrame
            a pandas dataframe with a geometry column
        cols : string or list of string
            name or list of names of columns to use to compute the statistic
        w : pysal weights object
            a weights object aligned with the dataframe. If not provided, this
            is searched for in the dataframe's metadata
        inplace : bool
            a boolean denoting whether to operate on the dataframe inplace or to
            return a series contaning the results of the computation. If
            operating inplace, the derived columns will be named 'column_g'
        pvalue : string
            a string denoting which pvalue should be returned. Refer to the
            the G statistic's documentation for available p-values
        outvals : list of strings
            list of arbitrary attributes to return as columns from the G statistic
        **stat_kws : dict
            options to pass to the underlying statistic. For this, see the
            documentation for the G statistic.

        Returns
        --------
        If inplace, None, and operation is conducted on dataframe
        in memory. Otherwise, returns a copy of the dataframe with
        the relevant columns attached.

        """

        msg = (
            "The `.by_cols()` methods are deprecated and will be "
            "removed in a future version of `esda`."
        )
        warnings.warn(msg, FutureWarning, stacklevel=2)

        return _univariate_handler(
            df,
            cols,
            w=w,
            inplace=inplace,
            pvalue=pvalue,
            outvals=outvals,
            stat=cls,
            swapname=cls.__name__.lower(),
            **stat_kws,
        )


class G_Local:  # noqa: N801
    """
    Generalized Local G Autocorrelation

    Parameters
    ----------
    y : array
        variable
    w : W | Graph
        spatial weights instance as W or Graph aligned with y
    transform : {'R', 'B'}
        the type of w, either 'B' (binary) or 'R' (row-standardized)
    permutations : int
        the number of random permutations for calculating
        pseudo p values
    star : boolean or float
        whether or not to include focal observation in sums (default: False)
        if the row-transformed weight is provided, then this is the default
        value to use within the spatial lag. Generally, weights should be
        provided in binary form, and standardization/self-weighting will be
        handled by the function itself.
    island_weight:
        value to use as a weight for the "fake" neighbor for every island.
        If numpy.nan, will propagate to the final local statistic depending
        on the `stat_func`. If 0, then the lag is always zero for islands.

    Attributes
    ----------
    y : array
       original variable
    w : W | Graph
        spatial weights instance as W or Graph aligned with y
    permutations : int
                  the number of permutations
    Gs : array
        of floats, the value of the orginal G statistic in Getis & Ord (1992)
    EGs : float
         expected value of Gs under normality assumption
         the values is scalar, since the expectation is identical
         across all observations
    VGs : array
         of floats, variance values of Gs under normality assumption
    Zs : array
        of floats, standardized Gs
    p_norm : array
            of floats, p-value under normality assumption (one-sided)
            for two-sided tests, this value should be multiplied by 2
    sim : array
         of arrays of floats (if permutations>0), vector of I values
         for permutated samples
    p_sim : array
        of floats, p-value based on permutations (one-sided)
        null - spatial randomness
        alternative - the observed G is extreme
            (it is either extremely high or extremely low)
    EG_sim : array
            of floats, average value of G from permutations
    VG_sim : array
            of floats, variance of G from permutations
    seG_sim : array
             of floats, standard deviation of G under permutations.
    z_sim : array
           of floats, standardized G based on permutations
    p_z_sim : array
             of floats, p-value based on standard normal approximation from
             permutations (one-sided)

    Notes
    -----
    To compute moments of Gs under normality assumption,
    PySAL considers w is either binary or row-standardized.
    For binary weights object, the weight value for self is 1
    For row-standardized weights object, the weight value for self is
    1/(the number of its neighbors + 1).


    For technical details see :cite:`Getis_2010` and :cite:`Ord_2010`.


    Examples
    --------
    >>> import libpysal
    >>> import numpy
    >>> numpy.random.seed(10)

    Preparing a point data set

    >>> points = [(10, 10), (20, 10), (40, 10), (15, 20), (30, 20), (30, 30)]

    Creating a weights object from points

    >>> w = libpysal.weights.DistanceBand(points,threshold=15)

    Preparing a variable

    >>> y = numpy.array([2, 3, 3.2, 5, 8, 7])

    Applying Getis and Ord local G test using a binary weights object

    >>> from esda.getisord import G_Local
    >>> lg = G_Local(y,w,transform='B')

    Examining the results

    >>> lg.Zs
    array([-1.0136729 , -0.04361589,  1.31558703, -0.31412676,  1.15373986,
            1.77833941])
    >>> round(lg.p_sim[0], 3)
    0.101

    p-value based on standard normal approximation from permutations
    >>> round(lg.p_z_sim[0], 3)
    0.154

    >>> numpy.random.seed(10)

    Applying Getis and Ord local G* test using a binary weights object

    >>> lg_star = G_Local(y,w,transform='B',star=True)

    Examining the results

    >>> lg_star.Zs
    array([-1.39727626, -0.28917762,  0.65064964, -0.28917762,  1.23452088,
            2.02424331])
    >>> round(lg_star.p_sim[0], 3)
    0.101

    >>> numpy.random.seed(12345)

    Applying Getis and Ord local G test using a row-standardized weights object

    >>> lg = G_Local(y,w,transform='R')

    Examining the results

    >>> lg.Zs
    array([-0.62074534, -0.01780611,  1.31558703, -0.12824171,  0.28843496,
            1.77833941])
    >>> round(lg.p_sim[0], 3)
    0.103

    >>> numpy.random.seed(10)

    Applying Getis and Ord local G* test using a row-standardized weights object

    >>> lg_star = G_Local(y,w,transform='R',star=True)

    Examining the results

    >>> lg_star.Zs
    array([-0.62488094, -0.09144599,  0.41150696, -0.09144599,  0.24690418,
            1.28024388])
    >>> round(lg_star.p_sim[0], 3)
    0.101

    """

    def __init__(
        self,
        y,
        w,
        transform="R",
        permutations=PERMUTATIONS,
        star=False,
        keep_simulations=True,
        n_jobs=-1,
        seed=None,
        island_weight=0,
    ):
        y = np.asarray(y).flatten()
        self.n = len(y)
        self.y = y
        w, star = _infer_star_and_structure_w(w, star, transform)
        if isinstance(w, W):
            w.transform = transform
        else:
            w = w.transform(transform)
        self.w_transform = transform
        self.w = w
        self.permutations = permutations
        self.star = star
        self.calc()
        self.p_norm = stats.norm.sf(np.abs(self.Zs))
        if permutations:
            self.p_sim, self.rGs = _crand_plus(
                y,
                w,
                self.Gs,
                permutations,
                keep_simulations,
                n_jobs=n_jobs,
                stat_func=_g_local_star_crand if star else _g_local_crand,
                scaling=y.sum(),
                seed=seed,
                island_weight=island_weight,
            )
            if keep_simulations:
                self.sim = sim = self.rGs.T
                self.EG_sim = sim.mean(axis=0)
                self.seG_sim = sim.std(axis=0)
                self.VG_sim = self.seG_sim * self.seG_sim
                self.z_sim = (self.Gs - self.EG_sim) / self.seG_sim
                self.p_z_sim = stats.norm.sf(np.abs(self.z_sim))

    def __crand(self, keep_simulations):
        y = self.y
        if keep_simulations:
            rGs = np.zeros((self.n, self.permutations))
        larger = np.zeros((self.n,))
        n_1 = self.n - 1
        rid = list(range(n_1))
        prange = list(range(self.permutations))
        if isinstance(self.w, W):
            k = self.w.max_neighbors + 1
        else:
            k = self.w.cardinalities.max()
        rids = np.array([np.random.permutation(rid)[0:k] for i in prange])
        ids = np.arange(self.w.n)
        wc = self.__getCardinalities()
        if self.w_transform == "r":  # noqa: SIM108 -- keeping readability
            den = np.array(wc) + self.star
        else:
            den = np.ones(self.w.n)
        for i in range(self.w.n):
            idsi = ids[ids != i]
            np.random.shuffle(idsi)
            yi_star = y[i] * self.star
            wci = wc[i]
            rGs_i = (y[idsi[rids[:, 0:wci]]]).sum(1) + yi_star
            rGs_i = (np.array(rGs_i) / den[i]) / (self.y_sum - (1 - self.star) * y[i])
            if keep_simulations:
                rGs[i] = rGs_i
            larger[i] = (rGs_i >= self.Gs[i]).sum()
        if keep_simulations:
            self.rGs = rGs
        below = (self.permutations - larger) < larger
        larger[below] = self.permutations - larger[below]
        self.p_sim = (larger + 1) / (self.permutations + 1)

    def __getCardinalities(self):  # noqa: N802
        if isinstance(self.w, W):
            ido = self.w.id_order
            self.wc = np.array([self.w.cardinalities[ido[i]] for i in range(self.n)])
        else:
            self.wc = self.w.cardinalities.values
        return self.wc

    def calc(self):
        w = self.w
        W = w.sparse

        self.y_sum = self.y.sum()

        y = self.y
        remove_self = not self.star
        N = self.w.n - remove_self

        statistic = (W @ y) / (y.sum() - y * remove_self)

        # ----------------------------------------------------#
        # compute moments necessary for analytical inference  #
        # ----------------------------------------------------#

        empirical_mean = (y.sum() - y * remove_self) / N
        # variance looks complex, yes, but it obtains from E[x^2] - E[x]^2.
        # So, break it down to allow subtraction of the self-neighbor.
        mean_of_squares = ((y**2).sum() - (y**2) * remove_self) / N
        empirical_variance = mean_of_squares - empirical_mean**2

        # Since we have corrected the diagonal, this should work
        cardinality = np.asarray(W.sum(axis=1)).squeeze()
        expected_value = cardinality / N

        expected_variance = cardinality * (N - cardinality)
        expected_variance /= N - 1
        expected_variance *= 1 / N**2
        expected_variance *= empirical_variance / (empirical_mean**2)

        z_scores = (statistic - expected_value) / np.sqrt(expected_variance)

        self.Gs = statistic
        self.EGs = expected_value
        self.VGs = expected_variance
        self.Zs = z_scores

    @property
    def _statistic(self):
        """Standardized accessor for esda statistics"""
        return self.Gs

    @classmethod
    def by_col(
        cls, df, cols, w=None, inplace=False, pvalue="sim", outvals=None, **stat_kws
    ):
        """
        Function to compute a G_Local statistic on a dataframe

        Parameters
        ----------
        df : pandas.DataFrame
            a pandas dataframe with a geometry column
        cols : string or list of string
            name or list of names of columns to use to compute the statistic
        w : W | Graph
            spatial weights instance as W or Graph aligned with the dataframe. If not
            provided, this is searched for in the dataframe's metadata
        inplace : bool
            a boolean denoting whether to operate on the dataframe inplace or to
            return a series contaning the results of the computation. If
            operating inplace, the derived columns will be named 'column_g_local'
        pvalue : string
            a string denoting which pvalue should be returned. Refer to the
            the G_Local statistic's documentation for available p-values
        outvals : list of strings
            list of arbitrary attributes to return as columns from the
            G_Local statistic
        **stat_kws : dict
            options to pass to the underlying statistic. For this, see the
            documentation for the G_Local statistic.

        Returns
        -------
        pandas.DataFrame
                        If inplace, None, and operation is conducted on dataframe
                        in memory. Otherwise, returns a copy of the dataframe with
                        the relevant columns attached.

        """

        msg = (
            "The `.by_col()` methods are deprecated and will be "
            "removed in a future version of `esda`."
        )
        warnings.warn(msg, FutureWarning, stacklevel=2)

        return _univariate_handler(
            df,
            cols,
            w=w,
            inplace=inplace,
            pvalue=pvalue,
            outvals=outvals,
            stat=cls,
            swapname=cls.__name__.lower(),
            **stat_kws,
        )


def _infer_star_and_structure_w(weights, star, transform):
    assert transform.lower() in ("r", "b"), (
        f'Transforms must be binary "b" or row-standardized "r".Recieved: {transform}'
    )
    adj_matrix = weights.sparse
    diagonal = adj_matrix.diagonal()
    zero_diagonal = (diagonal == 0).all()

    # Gi has a zero diagonal, Gi* has a nonzero diagonal
    star = (not zero_diagonal) if star is None else star

    # Want zero diagonal but do not have it
    if (not zero_diagonal) & (star is False):
        if isinstance(weights, W):
            weights = fill_diagonal(weights, 0)
        else:
            weights = weights.assign_self_weight(0).eliminate_zeros()
    # Want nonzero diagonal and have it
    elif (not zero_diagonal) & (star is True):  # noqa: SIM114 -- keeping readability
        weights = weights
    # Want zero diagonal and have it
    elif zero_diagonal & (star is False):
        weights = weights
    # Want nonzero diagonal and do not have it
    elif zero_diagonal & (star is True):
        # if the input is binary or requested transform is binary,
        # set the diagonal to 1.
        star_warn = (
            "Gi* requested, but (a) weights are already row-standardized,"
            " (b) no weights are on the diagonal, and"
            " (c) no default value supplied to star. Assuming that the"
            " self-weight is equivalent to the maximum weight in the"
            " row. To use a different default (like, .5), set `star=.5`,"
            " or use libpysal.weights.fill_diagonal() to set the diagonal"
            " values of your weights matrix and use `star=None` in Gi_Local."
        )
        if isinstance(weights, W):
            if transform.lower() == "b" or weights.transform.lower() == "b":
                weights = fill_diagonal(weights, 1)

            # if we know the target is row-standardized, use the row max
            # this works successfully for effectively binary but "O"-transformed input
            elif transform.lower() == "r":
                # This warning is presented in the documentation as well
                warnings.warn(star_warn, UserWarning, stacklevel=2)
                weights = fill_diagonal(
                    weights, np.asarray(adj_matrix.max(axis=1).todense()).flatten()
                )
        else:
            if transform.lower() == "b" or weights.transformation.lower() == "b":
                weights = weights.assign_self_weight(1)
            elif transform.lower() == "r":
                warnings.warn(star_warn, UserWarning, stacklevel=2)
                weights = weights.assign_self_weight(
                    np.asarray(adj_matrix.max(axis=1).todense()).flatten()
                )
    else:  # star was something else, so try to fill the weights with it
        try:
            if isinstance(weights, W):
                weights = fill_diagonal(weights, star)
            else:
                weights = weights.assign_self_weight(star)
        except TypeError:
            raise TypeError(
                f"Type of star ({type(star)}) not understood."
                f" Must be an integer, boolean, float, or numpy.ndarray."
            ) from None
    star = (weights.sparse.diagonal() > 0).any()
    if isinstance(weights, W):
        weights.transform = transform
    else:
        weights = weights.transform(transform)

    return weights, star


# --------------------------------------------------------------
# Conditional Randomization Function Implementations
# --------------------------------------------------------------


@_njit(fastmath=True)
def _g_local_crand(i, z, permuted_ids, weights_i, scaling):
    other_weights = weights_i[1:]
    zi, zrand = _prepare_univariate(i, z, permuted_ids, other_weights)
    return (zrand @ other_weights) / (scaling - zi)


@_njit(fastmath=True)
def _g_local_star_crand(i, z, permuted_ids, weights_i, scaling):
    self_weight = weights_i[0]
    other_weights = weights_i[1:]
    zi, zrand = _prepare_univariate(i, z, permuted_ids, other_weights)
    return (zrand @ other_weights + self_weight * zi) / scaling


if __name__ == "__main__":
    import geopandas
    import numpy
    from libpysal import examples, weights

    import esda

    df = geopandas.read_file(examples.get_path("NAT.shp"))

    w = weights.Rook.from_dataframe(df)

    for transform in ("r", "b"):
        for star in (True, False):
            test = esda.getisord.G_Local(df.GI89, w, transform=transform, star=star)
            out = test._calc2()
            (
                statistic,
                expected_value,
                expected_variance,
                z_scores,
                empirical_mean,
                empirical_variance,
            ) = out

            numpy.testing.assert_allclose(statistic, test.Gs)
            numpy.testing.assert_allclose(expected_value, test.EGs)
            numpy.testing.assert_allclose(expected_variance, test.VGs)
            numpy.testing.assert_allclose(z_scores, test.Zs)
            numpy.testing.assert_allclose(empirical_mean, test.yl_mean)
            numpy.testing.assert_allclose(empirical_variance, test.s2)

    # Also check that the None configuration works
    test = esda.getisord.G_Local(df.GI89, w, star=None)



import numpy as np
import pandas as pd
from libpysal import weights
from packaging.version import Version
from scipy import stats
from sklearn import __version__ as sklearn_version
from sklearn.base import BaseEstimator
from sklearn.utils import check_array

SKL_GE_16 = Version(sklearn_version) >= Version("1.6")


class Geary_Local_MV(BaseEstimator):  # noqa: N801
    """Local Geary - Multivariate"""

    def __init__(self, connectivity=None, permutations=999, drop_islands=True):
        """
        Initialize a Local_Geary_MV estimator

        Parameters
        ----------
        connectivity     : W | Graph
                           spatial weights instance as W or Graph aligned with y
        permutations     : int
                           (default=999)
                           number of random permutations for calculation
                           of pseudo p_values
        drop_islands : bool (default True)
            Whether or not to preserve islands as entries in the adjacency
            list. By default, observations with no neighbors do not appear
            in the adjacency list. If islands are kept, they are coded as
            self-neighbors with zero weight. See ``libpysal.weights.to_adjlist()``.

        Attributes
        ----------
        localG          : numpy array
                          array containing the observed multivariate
                          Local Geary values.
        p_sim           : numpy array
                          array containing the simulated
                          p-values for each unit.
        """

        self.connectivity = connectivity
        self.permutations = permutations
        self.drop_islands = drop_islands

    def fit(self, variables):
        """
        Parameters
        ----------
        variables        : numpy.ndarray
                           array containing continuous data

        Returns
        -------
        the fitted estimator.

        Notes
        -----
        Technical details and derivations can be found in :cite:`Anselin1995`.

        Examples
        --------
        Guerry data replication GeoDa tutorial
        >>> import libpysal
        >>> import geopandas as gpd
        >>> guerry = lp.examples.load_example('Guerry')
        >>> guerry_ds = gpd.read_file(guerry.get_path('Guerry.shp'))
        >>> w = libpysal.weights.Queen.from_dataframe(guerry_ds)
        >>> import libpysal
        >>> import geopandas as gpd
        >>> guerry = lp.examples.load_example('Guerry')
        >>> guerry_ds = gpd.read_file(guerry.get_path('Guerry.shp'))
        >>> w = libpysal.weights.Queen.from_dataframe(guerry_ds)
        >>> x1 = guerry_ds['Donatns']
        >>> x2 = guerry_ds['Suicids']
        >>> lG_mv = Local_Geary(connectivity=w).fit([x1,x2])
        >>> lG_mv.localG[0:5]
        >>> lG_mv.p_sim[0:5]
        """

        all_finite_kwarg = "ensure_all_finite" if SKL_GE_16 else "force_all_finite"

        self.variables = check_array(
            variables,
            accept_sparse=False,
            dtype="float",
            estimator=self,
            **{all_finite_kwarg: True},
        )

        w = self.connectivity
        if isinstance(w, weights.W):
            w.transform = "r"
        else:
            w = w.transform("r")

        self.n = len(variables[0])
        self.w = w

        permutations = self.permutations

        # Caclulate z-scores for input variables
        # to be used in _statistic and _crand
        zvariables = stats.zscore(variables, axis=1)

        stat = self._statistic_w if isinstance(w, weights.W) else self._statistic_g

        self.localG = stat(variables, zvariables, w, self.drop_islands)

        if permutations:
            self._crand(zvariables)
            sim = np.transpose(self.Gs)
            above = sim >= self.localG
            larger = above.sum(0)
            low_extreme = (permutations - larger) < larger
            larger[low_extreme] = permutations - larger[low_extreme]
            self.p_sim = (larger + 1.0) / (permutations + 1.0)

        return self

    @staticmethod
    def _stat(zseries, adj_list, variables):
        focal = [zseries[i].loc[adj_list.focal].values for i in range(len(variables))]
        neighbor = [
            zseries[i].loc[adj_list.neighbor].values for i in range(len(variables))
        ]
        # Carry out local Geary calculation
        gs = adj_list.weight.values * (np.array(focal) - np.array(neighbor)) ** 2
        return gs

    def _statistic_w(self, variables, zvariables, w, drop_islands):
        # Define denominator adjustment
        k = len(variables)
        # Create focal and neighbor values
        adj_list = w.to_adjlist(remove_symmetric=False, drop_islands=drop_islands)

        zseries = [pd.Series(i, index=w.id_order) for i in zvariables]
        gs = self._stat(zseries, adj_list, variables)

        # Reorganize data
        temp = pd.DataFrame(gs).T
        adj_list_gs = temp.groupby(by=adj_list.focal.values).sum()
        # Rearrange data based on w id order
        adj_list_gs["w_order"] = w.id_order
        adj_list_gs.sort_values(by="w_order", inplace=True)
        adj_list_gs.drop(columns=["w_order"], inplace=True)
        localG = np.array(adj_list_gs.sum(axis=1, numeric_only=True) / k)

        return localG

    def _statistic_g(self, variables, zvariables, w, drop_islands):
        # Define denominator adjustment
        k = len(variables)
        # Create focal and neighbor values
        if drop_islands:
            adj_list = w.adjacency.drop(w.isolates).reset_index()
        else:
            adj_list = w.adjacency.reset_index()
        zseries = [pd.Series(i, index=w.unique_ids) for i in zvariables]

        gs = self._stat(zseries, adj_list, variables)

        # Reorganize data
        temp = pd.DataFrame(gs).T
        adj_list_gs = temp.groupby(by=adj_list.focal.values).sum().reindex(w.unique_ids)
        localG = np.array(adj_list_gs.sum(axis=1, numeric_only=True) / k)

        return localG

    def _crand(self, zvariables):
        """
        conditional randomization

        for observation i with ni neighbors,  the candidate set cannot include
        i (we don't want i being a neighbor of i). we have to sample without
        replacement from a set of ids that doesn't include i. numpy doesn't
        directly support sampling wo replacement and it is expensive to
        implement this. instead we omit i from the original ids,  permute the
        ids and take the first ni elements of the permuted ids as the
        neighbors to i in each randomization.

        """
        nvars = self.variables.shape[0]
        Gs = np.zeros((self.n, self.permutations))
        prange = list(range(self.permutations))
        k = (
            self.w.max_neighbors + 1
            if isinstance(self.w, weights.W)
            else self.w.cardinalities.max() + 1
        )
        nn = self.n - 1
        rids = np.array([np.random.permutation(nn)[0:k] for i in prange])
        ids = np.arange(self.w.n)
        if isinstance(self.w, weights.W):
            ido = self.w.id_order
            w = [self.w.weights[ido[i]] for i in ids]
            wc = [self.w.cardinalities[ido[i]] for i in ids]
        else:
            w = list(self.w.weights.values())
            wc = self.w.cardinalities.tolist()

        for i in range(self.w.n):
            idsi = ids[ids != i]
            np.random.shuffle(idsi)
            vars_rand = []
            for j in range(nvars):
                vars_rand.append(zvariables[j][idsi[rids[:, 0 : wc[i]]]])
            # vars rand as tmp
            # Calculate diff
            diff = []
            for z in range(nvars):
                _diff = np.array((zvariables[z][i] - vars_rand[z]) ** 2 * w[i])
                diff.append(_diff.sum(1) / nvars)
            # add up differences
            temp = np.array([sum(x) for x in zip(*diff, strict=True)])
            # Assign to object to be returned
            Gs[i] = temp
        self.Gs = Gs



import libpysal
import numpy as np
import pytest

from esda.geary_local_mv import Geary_Local_MV

parametrize_w = pytest.mark.parametrize(
    "w",
    [
        libpysal.io.open(libpysal.examples.get_path("stl.gal")).read(),
        libpysal.graph.Graph.from_W(
            libpysal.io.open(libpysal.examples.get_path("stl.gal")).read()
        ),
    ],
    ids=["W", "Graph"],
)


class TestGearyLocalMV:
    def setup_method(self):
        np.random.seed(100)
        f = libpysal.io.open(libpysal.examples.get_path("stl_hom.txt"))
        self.y1 = np.array(f.by_col["HR8893"])
        self.y2 = np.array(f.by_col["HC8488"])

    @parametrize_w
    def test_defaults(self, w):
        lG_mv = Geary_Local_MV(connectivity=w).fit([self.y1, self.y2])
        print(lG_mv.p_sim[0])
        np.testing.assert_allclose(lG_mv.localG[0], 0.4096931479581422)
        np.testing.assert_allclose(lG_mv.p_sim[0], 0.211)



"""
:mod:`esda` --- Exploratory Spatial Data Analysis
=================================================

"""

import contextlib
from importlib.metadata import PackageNotFoundError, version

from . import adbscan, shape  # noqa F401
from .correlogram import correlogram  # noqa F401
from .gamma import Gamma  # noqa F401
from .geary import Geary  # noqa F401
from .geary_local import Geary_Local  # noqa F401
from .geary_local_mv import Geary_Local_MV  # noqa F401
from .getisord import G, G_Local  # noqa F401
from .join_counts import Join_Counts  # noqa F401
from .join_counts_local import Join_Counts_Local  # noqa F401
from .join_counts_local_bv import Join_Counts_Local_BV  # noqa F401
from .join_counts_local_mv import Join_Counts_Local_MV  # noqa F401
from .lee import Spatial_Pearson, Spatial_Pearson_Local  # noqa F401
from .losh import LOSH  # noqa F401
from .map_comparison import (
    areal_entropy,
    completeness,  # noqa F401
    external_entropy,
    homogeneity,
    overlay_entropy,
)
from .moran import (
    Moran,
    Moran_BV,
    Moran_BV_matrix,
    Moran_Local,  # noqa F401
    Moran_Local_BV,
    Moran_Local_Rate,
    Moran_Rate,
    plot_moran_facet,
)
from .moran_local_mv import (
    MoranLocalPartial,
    MoranLocalConditional
)
from .silhouettes import boundary_silhouette, path_silhouette  # noqa F401
from .smaup import Smaup  # noqa F401
from .topo import isolation, prominence  # noqa F401
from .util import fdr  # noqa F401

with contextlib.suppress(PackageNotFoundError):
    __version__ = version("esda")



import numpy as np
from .moran import Moran_Local
from .significance import calculate_significance
from libpysal.weights import lag_spatial
from libpysal.graph import Graph

try:
    from tqdm.auto import tqdm
except ImportError:

    def tqdm(x, **kwargs):
        return x


def _calc_quad(x,y):
    """
    This is a simpler solution to calculate a cartesian quadrant. 

    To explain graphically, let the tuple below be (off_sign[i], neg_y[i]*2). 

    If sign(x[i]) != sign(y[i]), we are on the negative diagonal. 
    If y is negative, we are on the bottom of the plot. 

    Therefore, the sum (off_sign + neg_y*2 + 1) gives you the cartesian quadrant. 

     II  |   I
     1,0 | 0,0
    -----+-----
     0,2 | 1,2
     III |  IV

    """
    off_sign = np.sign(x) != np.sign(y)
    neg_y = (y<0)
    return off_sign + neg_y*2 + 1

class MoranLocalPartial(object):
    def __init__(
        self, permutations=999, unit_scale=True, partial_labels=True, alternative='two-sided'
    ):
        """
        Compute the Multivariable Local Moran statistics under partial dependence :cite:`wolf2024confounded`

        Parameters
        ---------
        permutations : int
            the number of permutations to run for the inference,
            driven by conditional randomization.
        unit_scale : bool
            whether to enforce unit variance in the local statistics. This
            normalizes the variance of the data at inupt, ensuring that
            the covariance statistics are not overwhelmed by any single
            covariate's large variance.
        partial_labels : bool, default=True
            whether to calculate the classification based on the part-regressive 
            quadrant classification or the univariate quadrant classification,
            like a classical Moran's I. When mvquads is True, the variables are labelled as:
            - label 1: observations with large y - rho * x that also have large Wy values. 
            - label 2: observations with small y - rho * x values that also have large Wy values.
            - label 3: observations with small y - rho * x values that also have small Wy values. 
            - label 4: observations with large y - rho * x values that have small Wy values.
        alternative : str (default: 'two-sided')
            the alternative hypothesis for the inference. One of
            'two-sided', 'greater', 'lesser', 'directed', or 'folded'.
            See the esda.significance.calculate_significance() documentation
            for more information.

        Attributes
        ----------
        connectivity : The weights matrix inputted, but row standardized
        D : The "design" matrix used in computation. If X is
            not None, this will be [1 y X]
        R : the "response" matrix used in computation. Will
            always be the same shape as D and contain [1, Wy, Wy, ....]
        DtDi : empirical parameter covariance matrix
            the P x P matrix describing the variance and covariance
            of y and X.
        P : the number of parameters. 1 if X is not provided.
        association_ : the N,P matrix of multivariable LISA statistics.
            the first column, lmos[:,1] is the LISAs corresponding
            to the relationship between Wy and y conditioning on X.
        reference_distribution_ : the (N, permutations, P+1) realizations from the conditional
            randomization to generate reference distributions for
            each Local Moran statistic. rlmos_[:,:,1] pertain to
            the reference distribution of y and Wy.
        significance_  : the (N, P) matrix of quadrant classifications for the
            part-regressive relationships. quads[:,0] pertains to
            the relationship between y and Wy. The mean is not classified,
            since it's just binary above/below mean usually.
        partials_: the (N,2,P+1) matrix of part-regressive contributions.
            The ith slice of partials_[:,:,i] contains the
            partial regressive contribution of that covariate, with
            the first column indicating the part-regressive outcome
            and the second indicating the part-regressive design.
            The partial regression matrix starts at zero, so
            partials_[:,:,0] corresponds to the partial regression
            describing the relationship between y and Wy.
        labels_ : the (N,) array of quadrant classifications for the
            part-regressive relationships. See the partial_labels argument
            for more information. 
        """
        self.permutations = permutations
        self.unit_scale = unit_scale
        self.partial_labels = partial_labels
        self.alternative = alternative
    
    def fit(self, X, y, W):
        """
        Fit the partial local Moran statistic on input data

        Parameters
        ----------
        X   : (N,p) array
            array of data that is used as "confounding factors"
            to account for their covariance with Y.
        y   : (N,1) array
            array of data that is the targeted "outcome" covariate
            to compute the multivariable Moran's I
        W   : (N,N) weights object
            spatial weights instance as W or Graph aligned with y. Immediately row-standardized.
      
        Returns
        -------
        self    :   object
            this MoranLocalPartial() statistic after fitting to data
        """
        y = np.asarray(y).reshape(-1, 1)
        if isinstance(W, Graph):
            W = W.transform("R")
        else:
            W.transform = "r"
        y = y - y.mean()
        if self.unit_scale:
            y /= y.std()
        X = X - X.mean(axis=0)
        if self.unit_scale:
            X = X / X.std(axis=0)
        self.y = y
        self.X = X
        D, R = self._make_data(y, X, W)
        self.D, self.R = D, R
        self.P = D.shape[1] - 1
        self.N = W.n
        self.DtDi = np.linalg.inv(
            self.D.T @ self.D
        )  # this is only PxP, so not too bad...
        self._left_component_ = (self.D @ self.DtDi) * (self.N - 1)
        self._lmos_ = self._left_component_ * self.R
        self.connectivity = W
        self.permutations = self.permutations
        if self.permutations is not None:  # NOQA necessary to avoid None > 0
            if self.permutations > 0:
                self._crand(y, X, W)
                self._rlmos_ *= self.N - 1
                self._p_sim_ = np.zeros((self.N, self.P + 1))
                for i in range(self.P + 1):
                    self._p_sim_[:,i] = calculate_significance(
                        self._lmos_[:,i], 
                        self._rlmos_[:,:,i], 
                        alternative=self.alternative
                    )

        component_quads = []
        for i, left in enumerate(self._left_component_.T):
            right = self.R[:, i]
            quads = _calc_quad(left - left.mean(), right)
            component_quads.append(quads)
        self._partials_ = np.asarray(
            [
                np.vstack((left, right)).T
                for left, right in zip(self._left_component_.T, self.R.T)
            ]
        )

        uvquads = []
        negative_lag = R[:,1] < 0
        for i, x_ in enumerate(self.D.T):
            if i == 0:
                continue
            off_sign = np.sign(x_) != np.sign(R[:,1])
            quads = negative_lag.astype(int).flatten() * 2 + off_sign.astype(int) + 1
            uvquads.append(quads.flatten())

        self._uvquads_ = np.row_stack(uvquads).T
        self._mvquads_ = np.row_stack(component_quads).T
        return self

    def _make_data(self, z, X, W):
        if isinstance(W, Graph): # NOQA because ternary is confusing
            Wz = W.lag(z)
        else:
            Wz = lag_spatial(W, z)
        if X is not None:
            D = np.hstack((np.ones(z.shape), z, X))
            P = X.shape[1] + 1
        else:
            D = np.hstack((np.ones(z.shape), z))
            P = 1
        R = np.tile(Wz, P + 1)
        return D, R
        # self.D, self.R = D, R

    def _crand(self, y, X, W):
        N = W.n
        N_permutations = self.permutations
        prange = range(N_permutations)
        if isinstance(W, Graph):
            max_neighbs = W.cardinalities.max() + 1
        else:
            max_neighbs = W.max_neighbors + 1
        pre_permutations = np.array(
            [np.random.permutation(N - 1)[0:max_neighbs] for i in prange]
        )
        straight_ids = np.arange(N)
        if isinstance(W, Graph): # NOQA
            id_order = W.unique_ids
        else:
            id_order = W.id_order
        DtDi = self.DtDi
        ordered_weights = [W.weights[id_order[i]] for i in straight_ids]
        ordered_cardinalities = [W.cardinalities[id_order[i]] for i in straight_ids]
        lmos = np.empty((N, N_permutations, self.P + 1))
        for i in tqdm(range(N), desc="Simulating by site"):
            ids_noti = straight_ids[straight_ids != i]
            np.random.shuffle(ids_noti)
            these_permutations = pre_permutations[:, 0 : ordered_cardinalities[i]]
            randomized_permutations = ids_noti[these_permutations]
            shuffled_ys = y[randomized_permutations]
            these_weights = np.asarray(ordered_weights[i]).reshape(-1, 1)
            shuffled_Wyi = (shuffled_ys * these_weights).sum(
                axis=1
            )  # these are N-permutations by 1 now
            # shuffled_X = X[randomized_permutations, :] 
            # #these are still N-permutations, N-neighbs, N-covariates
            if X is None:
                local_data = np.array((1, y[i].item())).reshape(1, -1)
                shuffled_D = np.tile(
                    local_data, N_permutations
                ).T  # now N permutations by P
            else:
                local_data = np.array((1, y[i].item(), *X[i])).reshape(-1, 1)
                shuffled_D = np.tile(
                    local_data, N_permutations
                ).T  # now N permutations by P
            shuffled_R = np.tile(shuffled_Wyi, self.P + 1)
            lmos[i] = (shuffled_R * shuffled_D) @ DtDi
        self._rlmos_ = lmos  # nobs, nperm, nvars

    @property
    def association_(self):
        """
        The association between y and the local average of y,
        removing the correlation due to x and the local average of y
        """
        return self._lmos_[:, 1]

    @property
    def significance_(self):
        """
        The pseudo-p-value built using map randomization for the
        structural relationship between y and its local average,
        removing the correlation due to the relationship between x
        and the local average of y.
        """
        return self._p_sim_[:, 1]

    @property
    def partials_(self):
        """
        The components of the local statistic. The first column is the
        structural exogenous component of the data, and the second is the
        local average of y.
        """
        return self._partials_[1]

    @property
    def reference_distribution_(self):
        """
        Simulated distribution of association_, assuming that there is
          - no structural relationship between y and its local average;
          - the same observed structural relationship between y and x.
        """
        return self._rlmos_[:, :, 1]

    @property
    def labels_(self):
        """
        The classifications (in terms of cluster-type and outlier-type)
        for the association_ statistics. If the quads requested are
        *mvquads*, then the classification is done with respect to the
        left and right components (first and second columns of partials_).

        If the quads requested are *uvquads*, then this will only be computed
        with respect to the outcome and the local average.
        The cluster typology is:
          - 1: above-average left component (either y or D @ DtDi),
               above-average right component (local average of y)
          - 2: below-average left component (either y or D @ DtDi),
               above-average right component (local average of y)
          - 3: below-average left component (either y or D @ DtDi)
               below-average right component (local average of y)
          - 4: above-average left component (either y or D @ DtDi)
               below-average right component (local average of y)
        """
        if self.partial_labels:
            return self._mvquads_[:, 1]
        else:
            return self._uvquads_[:, 1]


class MoranLocalConditional(Moran_Local):
    """
    Fit a local moran statistic for y after regressing out the
    effects of confounding X on y. A "stronger" version of the
    MoranLocalPartial statistic, as defined by :cite:`wolf2024confounded`
    """

    def __init__(
        self,
        permutations=999,
        unit_scale=True,
        transformer=None,
        alternative='two-sided'
    ):
        """
        Initialize a local Moran statistic on the regression residuals

        Parameters
        ---------
        permutations    : int (default: 999)
                          the number of permutations to run for the inference,
                          driven by conditional randomization.
        unit_scale      : bool (default: True)
                          whether or not to convert the input data to a unit normal scale.
        transformer     : callable (default: scikit regression)
                          should transform X into a predicted y. If not provided, will use
                          the standard scikit OLS regression of y on X.
        alternative     : str (default: 'two-sided')
                          the alternative hypothesis for the inference. One of
                          'two-sided', 'greater', 'lesser', 'directed', or 'folded'.
                          See the esda.significance.calculate_significance() documentation
                          for more information.

        Attributes
        ----------
        connectivity : The weights matrix inputted, but row standardized
        y_filtered_ : the (N,1) array of y after removing the effect of X
        association_ : the N,P matrix of multivariable LISA statistics.
            the first column, lmos[:,1] is the LISAs corresponding
            to the relationship between Wy and y conditioning on X.
        reference_distribution_ : the (N, permutations, P+1) realizations from the conditional
            randomization to generate reference distributions for
            each Local Moran statistic. rlmos_[:,:,1] pertain to
            the reference distribution of y and Wy.
        significance_ : the (N, P) matrix of quadrant classifications for the
            part-regressive relationships. quads[:,0] pertains to
            the relationship between y and Wy. The mean is not classified,
            since it's just binary above/below mean usually.
        partials_ : the (N,2,P+1) matrix of part-regressive contributions.
            The ith slice of partials_[:,:,i] contains the
            partial regressive contribution of that covariate, with
            the first column indicating the part-regressive outcome
            and the second indicating the part-regressive design.
            The partial regression matrix starts at zero, so
            partials_[:,:,0] corresponds to the partial regression
            describing the relationship between y and Wy.
        labels_ : the (N,) array of quadrant classifications for the
            part-regressive relationships. See the partial_labels argument
            for more information. 
        """
        self.permutations = permutations
        self.unit_scale = unit_scale
        self.transformer = transformer
        self.alternative = alternative

    def fit(self, X, y, W):
        """
        Parameters
        ---------
        y : (N,1) array
            array of data that is the targeted "outcome" covariate
            to compute the multivariable Moran's I
        X : (N,3) array
            array of data that is used as "confounding factors"
            to account for their covariance with Y.
        W : (N,N) weights object
            spatial weights instance as W or Graph aligned with y. Immediately row-standardized.
        
        Returns
        -------
        A fitted MoranLocalConditional() estimator
        """
        y = y - y.mean()
        X = X - X.mean(axis=0)
        if self.unit_scale:
            y /= y.std()
            X /= X.std(axis=0)
        self.y = y
        self.X = X
        y_filtered_ = self.y_filtered_ = self._part_regress_transform(y, X)
        if isinstance(W, Graph):
            W = W.transform("R")
            Wyf = W.lag(y_filtered_)
        else:
            W.transform = "r"
            Wyf = lag_spatial(W, y_filtered_) # TODO: graph
        self.connectivity = W
        self.partials_ = np.column_stack((y_filtered_, Wyf))
        y_out = self.y_filtered_
        self.association_ = ((y_out * Wyf) / (y_out.T @ y_out) * (W.n - 1)).flatten()
        if self.permutations > 0:
            self._crand()
            self.significance_ = calculate_significance(self.association_, self.reference_distribution_, alternative=self.alternative)
        quads = np.array([[3,2,4,1]]).reshape(2,2)
        left_component_cluster = (y_filtered_ > 0).astype(int)
        right_component_cluster = (Wyf > 0).astype(int)
        quads = quads[left_component_cluster, right_component_cluster]
        self.labels_ = quads.squeeze()
        return self

    def _part_regress_transform(self, y, X):
        """If the object has a _transformer, use it; otherwise, fit it."""
        if hasattr(self, "_transformer"):
            ypart = y - self._transformer(X)
        else:
            from sklearn.linear_model import LinearRegression

            self._transformer = LinearRegression().fit(X, y).predict
            ypart = self._part_regress_transform(y, X)
        return ypart

    def _crand(self):
        """Cribbed from esda.Moran_Local
        conditional randomization
        for observation i with ni neighbors,  the candidate set cannot include
        i (we don't want i being a neighbor of i). we have to sample without
        replacement from a set of ids that doesn't include i. numpy doesn't
        directly support sampling wo replacement and it is expensive to
        implement this. instead we omit i from the original ids,  permute the
        ids and take the first ni elements of the permuted ids as the
        neighbors to i in each randomization.
        """
        _, z = self.partials_.T
        lisas = np.zeros((self.connectivity.n, self.permutations))
        n_1 = self.connectivity.n - 1
        prange = list(range(self.permutations))
        if isinstance(self.connectivity, Graph):
            k = self.connectivity.cardinalities.max() + 1
        else:
            k = self.connectivity.max_neighbors + 1
        nn = self.connectivity.n - 1
        rids = np.array([np.random.permutation(nn)[0:k] for i in prange])
        ids = np.arange(self.connectivity.n)
        if hasattr(self.connectivity, "id_order"):
            ido = self.connectivity.id_order
        else:
            ido = self.connectivity.unique_ids.values
        w = [self.connectivity.weights[ido[i]] for i in ids]
        wc = [self.connectivity.cardinalities[ido[i]] for i in ids]

        for i in tqdm(range(self.connectivity.n), desc="Simulating by site"):
            idsi = ids[ids != i]
            np.random.shuffle(idsi)
            tmp = z[idsi[rids[:, 0 : wc[i]]]]
            lisas[i] = z[i] * (w[i] * tmp).sum(1)
        self.reference_distribution_ = (n_1 / (z * z).sum()) * lisas


MoranLocalConditional.__init__.__doc__ = MoranLocalPartial.__init__.__doc__.replace(
    "Partial", "Conditional"
)



import warnings

import libpysal as lp
import numpy as np
from scipy import stats
from sklearn.base import BaseEstimator


def _slag(w, y):
    """Helper to compute lag either for W or for Graph"""
    if isinstance(w, lp.weights.W):
        return lp.weights.lag_spatial(w, y)
    else:
        return w.lag(y)


class LOSH(BaseEstimator):
    """Local spatial heteroscedasticity (LOSH)"""

    def __init__(self, connectivity=None, inference=None):
        """
        Initialize a losh estimator

        Parameters
        ----------
        connectivity     : W | Graph
                          spatial weights instance as W or Graph aligned with y
        inference        : str
                           describes type of inference to be used. options are
                           "chi-square" or "permutation" methods.

        Attributes
        ----------
        Hi               : numpy array
                           Array of LOSH values for each spatial unit.
        ylag             : numpy array
                           Spatially lagged y values.
        yresid           : numpy array
                           Spatially lagged residual values.
        VarHi            : numpy array
                           Variance of Hi.
        pval             : numpy array
                           P-values for inference based on either
                           "chi-square" or "permutation" methods.
        """

        self.connectivity = connectivity
        self.inference = inference

    def fit(self, y, a=2):
        """
        Parameters
        ----------
        y                : numpy.ndarray
                           array containing continuous data
        a                : int
                           residual multiplier. Default is 2 in order
                           to generate a variance measure. Users may
                           use 1 for absolute deviations.

        Returns
        -------
        the fitted estimator.

        Notes
        -----
        Technical details and derivations can be found in :cite:`OrdGetis2012`.

        Examples
        --------
        >>> import libpysal
        >>> w = libpysal.io.open(libpysal.examples.get_path("stl.gal")).read()
        >>> f = libpysal.io.open(libpysal.examples.get_path("stl_hom.txt"))
        >>> y = np.array(f.by_col['HR8893'])
        >>> from esda import losh
        >>> ls = losh(connectivity=w, inference="chi-square").fit(y)
        >>> np.round(ls.Hi[0], 3)
        >>> np.round(ls.pval[0], 3)

        Boston housing data replicating R spdep::LOSH()
        >>> import libpysal
        >>> import geopandas as gpd
        >>> boston = libpysal.examples.load_example('Bostonhsg')
        >>> boston_ds = gpd.read_file(boston.get_path('boston.shp'))
        >>> w = libpysal.weights.Queen.from_dataframe(boston_ds)
        >>> ls = losh(connectivity=w, inference="chi-square").fit(boston['NOX'])
        >>> np.round(ls.Hi[0], 3)
        >>> np.round(ls.VarHi[0], 3)
        """
        y = np.asarray(y).flatten()

        w = self.connectivity

        self.Hi, self.ylag, self.yresid, self.VarHi = self._statistic(y, w, a)

        if self.inference is None:
            return self
        elif self.inference == "chi-square":
            if a != 2:
                warnings.warn(
                    (
                        "Chi-square inference assumes that a=2, but "
                        f"a={a}. This means the inference will be invalid!"
                    ),
                    UserWarning,
                    stacklevel=2,
                )
            else:
                dof = 2 / self.VarHi
                Zi = (2 * self.Hi) / self.VarHi
                self.pval = 1 - stats.chi2.cdf(Zi, dof)
        else:
            raise NotImplementedError(
                "The requested inference method "
                f"({self.inference}) is not currently supported!"
            )

        return self

    @staticmethod
    def _statistic(y, w, a):
        # Define what type of variance to use
        a = 2 if a is None else a

        rowsum = np.array(w.sparse.sum(axis=1)).flatten()

        # Calculate spatial mean
        ylag = _slag(w, y) / rowsum
        # Calculate and adjust residuals based on multiplier
        yresid = abs(y - ylag) ** a
        # Calculate denominator of Hi equation
        denom = np.mean(yresid) * np.array(rowsum)
        # Carry out final Hi calculation
        Hi = _slag(w, yresid) / denom
        # Calculate average of residuals
        yresid_mean = np.mean(yresid)

        # Calculate VarHi
        n = len(y)
        squared_rowsum = np.asarray(w.sparse.multiply(w.sparse).sum(axis=1)).flatten()
        term1 = (n - 1) ** -1
        term2 = denom**-2
        term3 = (np.sum(yresid**2) / n) - yresid_mean**2
        term4 = (n * squared_rowsum) - (rowsum**2)
        VarHi = term1 * term2 * term3 * term4

        return (Hi, ylag, yresid, VarHi)
